"""Handle stream-level logic for the protocol engine."""

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

from aioquic._buffer import Buffer as QuicBuffer

from pywebtransport import constants
from pywebtransport._protocol.events import (
    Effect,
    EmitStreamEvent,
    InternalBindQuicStream,
    InternalFailQuicStream,
    InternalReturnStreamData,
    NotifyRequestDone,
    NotifyRequestFailed,
    ResetQuicStream,
    SendH3Capsule,
    SendQuicData,
    StopQuicStream,
    TransportStreamReset,
    UserGetStreamDiagnostics,
    UserResetStream,
    UserSendStreamData,
    UserStopStream,
    UserStreamRead,
    WebTransportStreamDataReceived,
)
from pywebtransport._protocol.state import ProtocolState, SessionStateData, StreamStateData
from pywebtransport._protocol.utils import (
    calculate_new_data_limit,
    calculate_new_stream_limit,
    get_stream_direction_from_id,
    http_code_to_webtransport_code,
    is_peer_initiated_stream,
    is_unidirectional_stream,
    webtransport_code_to_http_code,
)
from pywebtransport.constants import ErrorCodes
from pywebtransport.exceptions import SessionError, StreamError
from pywebtransport.types import Buffer, EventType, SessionState, StreamDirection, StreamState
from pywebtransport.utils import ensure_buffer, get_logger, get_timestamp

if TYPE_CHECKING:
    from pywebtransport.config import ClientConfig, ServerConfig


__all__: list[str] = []

logger = get_logger(name=__name__)

OPTIMIZED_READ_SLICE_THRESHOLD = 32 * 1024


class StreamProcessor:
    """Process stream-level events and manage state transitions."""

    def __init__(self, *, is_client: bool, config: ClientConfig | ServerConfig) -> None:
        """Initialize the stream processor."""
        self._is_client = is_client
        self._config = config

    def handle_get_stream_diagnostics(self, *, event: UserGetStreamDiagnostics, state: ProtocolState) -> list[Effect]:
        """Handle the UserGetStreamDiagnostics event."""
        stream_id = event.stream_id
        stream_data = state.streams.get(stream_id)
        if stream_data is None:
            return [
                NotifyRequestFailed(
                    request_id=event.request_id,
                    exception=StreamError(message=f"Stream {stream_id} not found for diagnostics", stream_id=stream_id),
                )
            ]
        data_dict = dataclasses.asdict(stream_data)
        data_dict["read_buffer"] = bytes(0)
        data_dict["read_buffer_size"] = stream_data.read_buffer_size
        return [NotifyRequestDone(request_id=event.request_id, result=data_dict)]

    def handle_internal_bind_quic_stream(self, *, event: InternalBindQuicStream, state: ProtocolState) -> list[Effect]:
        """Handle the InternalBindQuicStream event."""
        effects: list[Effect] = []
        session_id = event.session_id
        session_data = state.sessions.get(session_id)

        if session_data is None:
            effects.append(
                NotifyRequestFailed(
                    request_id=event.request_id,
                    exception=SessionError(f"Session {session_id} not found during stream bind"),
                )
            )
            return effects

        stream_id = event.stream_id
        direction = StreamDirection.SEND_ONLY if event.is_unidirectional else StreamDirection.BIDIRECTIONAL

        stream_data = StreamStateData(
            stream_id=stream_id,
            session_id=session_id,
            direction=direction,
            state=StreamState.OPEN,
            created_at=get_timestamp(),
        )

        state.streams[stream_id] = stream_data
        session_data.active_streams.add(stream_id)

        effects.append(NotifyRequestDone(request_id=event.request_id, result=stream_id))
        effects.append(
            EmitStreamEvent(
                stream_id=stream_id,
                event_type=EventType.STREAM_OPENED,
                data={"stream_id": stream_id, "session_id": session_id, "direction": direction},
            )
        )
        return effects

    def handle_internal_fail_quic_stream(self, *, event: InternalFailQuicStream, state: ProtocolState) -> list[Effect]:
        """Handle the InternalFailQuicStream event."""
        session_data = state.sessions.get(event.session_id)
        if session_data is not None:
            if event.is_unidirectional:
                if session_data.local_streams_uni_opened > 0:
                    session_data.local_streams_uni_opened -= 1
            else:
                if session_data.local_streams_bidi_opened > 0:
                    session_data.local_streams_bidi_opened -= 1

        return [NotifyRequestFailed(request_id=event.request_id, exception=event.exception)]

    def handle_return_stream_data(self, *, event: InternalReturnStreamData, state: ProtocolState) -> list[Effect]:
        """Handle the InternalReturnStreamData event."""
        stream_data = state.streams.get(event.stream_id)
        if stream_data is not None:
            stream_data.read_buffer.appendleft(event.data)
            stream_data.read_buffer_size += len(event.data)
        return []

    def handle_reset_stream(self, *, event: UserResetStream, state: ProtocolState) -> list[Effect]:
        """Handle the UserResetStream event to reset the sending side."""
        effects: list[Effect] = []
        stream_id = event.stream_id
        stream_data = state.streams.get(stream_id)

        if stream_data is None:
            error = StreamError(message=f"Stream {stream_id} not found for reset", stream_id=stream_id)
            effects.append(NotifyRequestFailed(request_id=event.request_id, exception=error))
            return effects

        if stream_data.state in (StreamState.HALF_CLOSED_LOCAL, StreamState.CLOSED, StreamState.RESET_SENT):
            effects.append(NotifyRequestDone(request_id=event.request_id, result=None))
            return effects

        original_state = stream_data.state

        stream_data.state = StreamState.RESET_SENT
        stream_data.closed_at = get_timestamp()
        stream_data.close_code = event.error_code

        http_error_code = webtransport_code_to_http_code(app_error_code=event.error_code)
        effects.append(ResetQuicStream(stream_id=stream_id, error_code=http_error_code))
        effects.append(NotifyRequestDone(request_id=event.request_id, result=None))

        while stream_data.write_buffer:
            _data_chunk, req_id, _end_stream = stream_data.write_buffer.popleft()
            error = StreamError(
                message=f"Stream {stream_id} reset by application", stream_id=stream_id, error_code=event.error_code
            )
            effects.append(NotifyRequestFailed(request_id=req_id, exception=error))
        stream_data.write_buffer_size = 0

        session_data = state.sessions.get(stream_data.session_id)
        if session_data is not None:
            session_data.blocked_streams.discard(stream_id)

        match original_state:
            case StreamState.HALF_CLOSED_REMOTE | StreamState.RESET_RECEIVED:
                stream_data.state = StreamState.CLOSED
                if session_data is not None:
                    session_data.active_streams.discard(stream_id)
                    effects.extend(self._process_stream_closure(session_data=session_data, stream_id=stream_id))
                effects.append(
                    EmitStreamEvent(
                        stream_id=stream_id, event_type=EventType.STREAM_CLOSED, data={"stream_id": stream_id}
                    )
                )

        logger.debug(
            "Stream %d reset locally with code %d (mapped to %x)", stream_id, event.error_code, http_error_code
        )
        return effects

    def handle_send_stream_data(self, *, event: UserSendStreamData, state: ProtocolState) -> list[Effect]:
        """Handle the UserSendStreamData event to send data."""
        effects: list[Effect] = []
        stream_id = event.stream_id
        stream_data = state.streams.get(stream_id)

        if stream_data is None:
            error = StreamError(message=f"Stream {stream_id} not found for sending data", stream_id=stream_id)
            effects.append(NotifyRequestFailed(request_id=event.request_id, exception=error))
            return effects

        if stream_data.state in (StreamState.HALF_CLOSED_LOCAL, StreamState.CLOSED, StreamState.RESET_SENT):
            error = StreamError(
                message=f"Stream {stream_id} is not writable (state: {stream_data.state})", stream_id=stream_id
            )
            effects.append(NotifyRequestFailed(request_id=event.request_id, exception=error))
            return effects

        session_id = stream_data.session_id
        session_data = state.sessions.get(session_id)
        if session_data is None:
            logger.error("Internal state error: Stream %d exists but session %s does not", stream_id, session_id)
            error = StreamError(message="Internal state error: Session not found", stream_id=stream_id)
            effects.append(NotifyRequestFailed(request_id=event.request_id, exception=error))
            return effects

        try:
            buffer_data = ensure_buffer(data=event.data)
        except TypeError as exc:
            logger.warning("Stream %d received invalid data type: %s", stream_id, exc)
            return [NotifyRequestFailed(request_id=event.request_id, exception=exc)]

        data_len = len(buffer_data)
        max_buffer_size = self._config.max_stream_write_buffer
        current_buffer_size = stream_data.write_buffer_size

        if current_buffer_size + data_len > max_buffer_size:
            error = StreamError(
                message=(
                    f"Stream {stream_id} write buffer full "
                    f"({current_buffer_size} + {data_len} > {max_buffer_size} bytes)"
                ),
                stream_id=stream_id,
            )
            return [NotifyRequestFailed(request_id=event.request_id, exception=error)]

        if stream_data.write_buffer:
            logger.debug("Stream %d write added to existing write buffer", stream_id)
            stream_data.write_buffer.append((buffer_data, event.request_id, event.end_stream))
            stream_data.write_buffer_size += data_len
            session_data.blocked_streams.add(stream_id)
            return []

        available_credit = session_data.peer_max_data - session_data.local_data_sent

        if data_len <= available_credit:
            session_data.local_data_sent += data_len
            stream_data.bytes_sent += data_len
            effects.append(SendQuicData(stream_id=stream_id, data=buffer_data, end_stream=event.end_stream))
            effects.append(NotifyRequestDone(request_id=event.request_id, result=None))

            if event.end_stream:
                original_state = stream_data.state
                match original_state:
                    case StreamState.HALF_CLOSED_REMOTE | StreamState.RESET_RECEIVED:
                        stream_data.state = StreamState.CLOSED
                        session_data.active_streams.discard(stream_id)
                        session_data.blocked_streams.discard(stream_id)
                        effects.extend(self._process_stream_closure(session_data=session_data, stream_id=stream_id))
                        effects.append(
                            EmitStreamEvent(
                                stream_id=stream_id, event_type=EventType.STREAM_CLOSED, data={"stream_id": stream_id}
                            )
                        )
                    case StreamState.OPEN:
                        stream_data.state = StreamState.HALF_CLOSED_LOCAL
                logger.debug("Stream %d send side closed", stream_id)

            return effects

        elif available_credit > 0:
            data_to_send_now = buffer_data[:available_credit]
            remaining_data = buffer_data[available_credit:]

            session_data.local_data_sent += available_credit
            stream_data.bytes_sent += available_credit
            effects.append(SendQuicData(stream_id=stream_id, data=data_to_send_now, end_stream=False))

            logger.debug(
                "Stream %d partial send: sent %d bytes, buffering %d bytes",
                stream_id,
                available_credit,
                len(remaining_data),
            )
            stream_data.write_buffer.append((remaining_data, event.request_id, event.end_stream))
            stream_data.write_buffer_size += len(remaining_data)
            session_data.blocked_streams.add(stream_id)

            buf = QuicBuffer(capacity=8)
            buf.push_uint_var(session_data.peer_max_data)
            effects.append(
                SendH3Capsule(
                    stream_id=session_data.session_id,
                    capsule_type=constants.WT_DATA_BLOCKED_TYPE,
                    capsule_data=buf.data,
                    end_stream=False,
                )
            )
            return effects

        else:
            logger.debug(
                "Stream %d write blocked by session flow control (%d > %d)", stream_id, data_len, available_credit
            )
            stream_data.write_buffer.append((buffer_data, event.request_id, event.end_stream))
            stream_data.write_buffer_size += data_len
            session_data.blocked_streams.add(stream_id)

            buf = QuicBuffer(capacity=8)
            buf.push_uint_var(session_data.peer_max_data)
            effects.append(
                SendH3Capsule(
                    stream_id=session_data.session_id,
                    capsule_type=constants.WT_DATA_BLOCKED_TYPE,
                    capsule_data=buf.data,
                    end_stream=False,
                )
            )
            return effects

    def handle_stop_stream(self, *, event: UserStopStream, state: ProtocolState) -> list[Effect]:
        """Handle the UserStopStream event to stop the receiving side."""
        effects: list[Effect] = []
        stream_id = event.stream_id
        stream_data = state.streams.get(stream_id)

        if stream_data is None:
            error = StreamError(message=f"Stream {stream_id} not found for stop", stream_id=stream_id)
            effects.append(NotifyRequestFailed(request_id=event.request_id, exception=error))
            return effects

        if stream_data.state in (StreamState.HALF_CLOSED_REMOTE, StreamState.CLOSED, StreamState.RESET_RECEIVED):
            effects.append(NotifyRequestDone(request_id=event.request_id, result=None))
            return effects

        original_state = stream_data.state
        stream_data.state = StreamState.RESET_RECEIVED
        stream_data.closed_at = get_timestamp()
        stream_data.close_code = event.error_code

        http_error_code = webtransport_code_to_http_code(app_error_code=event.error_code)
        effects.append(StopQuicStream(stream_id=stream_id, error_code=http_error_code))
        effects.append(NotifyRequestDone(request_id=event.request_id, result=None))

        while stream_data.pending_read_requests:
            req_id, _ = stream_data.pending_read_requests.popleft()
            error = StreamError(
                message=f"Stream {stream_id} stopped by application", stream_id=stream_id, error_code=event.error_code
            )
            effects.append(NotifyRequestFailed(request_id=req_id, exception=error))

        session_data = state.sessions.get(stream_data.session_id)

        match original_state:
            case StreamState.HALF_CLOSED_LOCAL | StreamState.RESET_SENT:
                stream_data.state = StreamState.CLOSED
                if session_data is not None:
                    session_data.active_streams.discard(stream_id)
                    session_data.blocked_streams.discard(stream_id)
                    effects.extend(self._process_stream_closure(session_data=session_data, stream_id=stream_id))
                effects.append(
                    EmitStreamEvent(
                        stream_id=stream_id, event_type=EventType.STREAM_CLOSED, data={"stream_id": stream_id}
                    )
                )

        logger.debug(
            "Stream %d receive side stopped locally with code %d (mapped to %x)",
            stream_id,
            event.error_code,
            http_error_code,
        )
        return effects

    def handle_stream_read(self, *, event: UserStreamRead, state: ProtocolState) -> list[Effect]:
        """Handle the UserStreamRead event to read data."""
        effects: list[Effect] = []
        stream_id = event.stream_id
        stream_data = state.streams.get(stream_id)

        if stream_data is None:
            error = StreamError(message=f"Stream {stream_id} not found for reading", stream_id=stream_id)
            effects.append(NotifyRequestFailed(request_id=event.request_id, exception=error))
            return effects

        if stream_data.read_buffer_size > 0:
            target = self._calculate_read_amount(
                requested_bytes=event.max_bytes, buffer_size=stream_data.read_buffer_size
            )
            data_to_return = self._read_from_buffer(stream_data=stream_data, max_bytes=target)
            effects.append(NotifyRequestDone(request_id=event.request_id, result=data_to_return))

            session_data = state.sessions.get(stream_data.session_id)
            if session_data is not None:
                session_data.local_data_consumed += len(data_to_return)
                credit_effect = self._check_and_send_data_credit(session_data=session_data)
                if credit_effect is not None:
                    effects.append(credit_effect)

            return effects

        if stream_data.state in (StreamState.RESET_RECEIVED, StreamState.CLOSED):
            error = StreamError(
                message=f"Stream {stream_id} receive side closed (state: {stream_data.state})", stream_id=stream_id
            )
            effects.append(NotifyRequestFailed(request_id=event.request_id, exception=error))
            return effects

        if stream_data.state == StreamState.HALF_CLOSED_REMOTE:
            effects.append(NotifyRequestDone(request_id=event.request_id, result=b""))
            return effects

        stream_data.pending_read_requests.append((event.request_id, event.max_bytes))
        return effects

    def handle_transport_stream_reset(self, *, event: TransportStreamReset, state: ProtocolState) -> list[Effect]:
        """Handle a transport-level stream reset from the peer."""
        effects: list[Effect] = []
        stream_id = event.stream_id
        stream_data = state.streams.get(stream_id)

        if stream_data is None or stream_data.state == StreamState.CLOSED:
            return []

        logger.debug("Stream %d reset by peer with code %d", stream_id, event.error_code)

        app_error_code = event.error_code
        if ErrorCodes.WT_APPLICATION_ERROR_FIRST <= event.error_code <= ErrorCodes.WT_APPLICATION_ERROR_LAST:
            try:
                app_error_code = http_code_to_webtransport_code(http_error_code=event.error_code)
            except ValueError:
                logger.warning(
                    "Received reserved H3 error code %x on stream %d, using as-is.", event.error_code, stream_id
                )

        stream_data.closed_at = get_timestamp()
        stream_data.close_code = app_error_code

        while stream_data.pending_read_requests:
            req_id, _ = stream_data.pending_read_requests.popleft()
            error = StreamError(
                message=f"Stream {stream_id} reset by peer", stream_id=stream_id, error_code=app_error_code
            )
            effects.append(NotifyRequestFailed(request_id=req_id, exception=error))

        while stream_data.write_buffer:
            _data_chunk, req_id, _end_stream = stream_data.write_buffer.popleft()
            error = StreamError(
                message=f"Stream {stream_id} reset by peer", stream_id=stream_id, error_code=app_error_code
            )
            effects.append(NotifyRequestFailed(request_id=req_id, exception=error))
        stream_data.write_buffer_size = 0

        stream_data.state = StreamState.CLOSED

        session_data = state.sessions.get(stream_data.session_id)
        if session_data is not None:
            session_data.active_streams.discard(stream_id)
            session_data.blocked_streams.discard(stream_id)
            effects.extend(self._process_stream_closure(session_data=session_data, stream_id=stream_id))

        effects.append(
            EmitStreamEvent(stream_id=stream_id, event_type=EventType.STREAM_CLOSED, data={"stream_id": stream_id})
        )

        return effects

    def handle_webtransport_stream_data(
        self, *, event: WebTransportStreamDataReceived, state: ProtocolState
    ) -> list[Effect]:
        """Handle WebTransport data received on an established stream."""
        effects: list[Effect] = []
        stream_id = event.stream_id
        stream_data = state.streams.get(stream_id)
        session_data: SessionStateData | None = None

        if stream_data is None:
            if self._is_client:
                logger.warning("Client received WT data for unknown stream %d, ignoring.", stream_id)
                return []

            session_id = event.session_id
            if state.sessions.get(session_id) is None:
                if state.early_event_count >= self._config.max_total_pending_events:
                    logger.warning(
                        "Global early event buffer full (%d), rejecting stream %d", state.early_event_count, stream_id
                    )
                    return [
                        ResetQuicStream(
                            stream_id=stream_id, error_code=constants.ErrorCodes.WT_BUFFERED_STREAM_REJECTED
                        )
                    ]

                session_buffer = state.early_event_buffer.get(session_id, [])
                if len(session_buffer) >= self._config.max_pending_events_per_session:
                    logger.warning(
                        "Per-session early event buffer full (%d) for session %d, rejecting stream %d",
                        len(session_buffer),
                        session_id,
                        stream_id,
                    )
                    return [
                        ResetQuicStream(
                            stream_id=stream_id, error_code=constants.ErrorCodes.WT_BUFFERED_STREAM_REJECTED
                        )
                    ]

                logger.debug("Buffering early event for stream %d on unknown session %d", stream_id, session_id)
                state.early_event_buffer.setdefault(session_id, []).append((get_timestamp(), event))
                state.early_event_count += 1
                return []

            session_data = state.sessions.get(session_id)
            if session_data is None:
                logger.warning(
                    "Received WT data for stream %d with unknown session %s, ignoring.", stream_id, session_id
                )
                return effects

            if session_data.state not in (SessionState.CONNECTED, SessionState.DRAINING):
                logger.debug(
                    "Ignoring new stream %d for session %s in state %s", stream_id, session_id, session_data.state
                )
                return []

            direction = get_stream_direction_from_id(stream_id=stream_id, is_client=self._is_client)

            match direction:
                case StreamDirection.RECEIVE_ONLY:
                    if session_data.peer_streams_uni_opened >= session_data.local_max_streams_uni:
                        logger.warning(
                            "Session %s unidirectional stream limit (%d) reached, ignoring stream %d",
                            session_id,
                            session_data.local_max_streams_uni,
                            stream_id,
                        )
                        return []
                    session_data.peer_streams_uni_opened += 1

                case StreamDirection.BIDIRECTIONAL:
                    if session_data.peer_streams_bidi_opened >= session_data.local_max_streams_bidi:
                        logger.warning(
                            "Session %s bidirectional stream limit (%d) reached, ignoring stream %d",
                            session_id,
                            session_data.local_max_streams_bidi,
                            stream_id,
                        )
                        return []
                    session_data.peer_streams_bidi_opened += 1

                case StreamDirection.SEND_ONLY:
                    logger.warning(
                        "Received WT data on server for client-initiated send-only stream %d, ignoring.", stream_id
                    )
                    return []

                case _:
                    raise AssertionError(f"Unreachable code: Unhandled stream direction {direction}")

            logger.debug("Creating new incoming stream %d for session %s", stream_id, session_id)
            stream_data = StreamStateData(
                stream_id=stream_id,
                session_id=session_id,
                direction=direction,
                state=StreamState.OPEN,
                created_at=get_timestamp(),
            )
            state.streams[stream_id] = stream_data
            session_data.active_streams.add(stream_id)

            effects.append(
                EmitStreamEvent(
                    stream_id=stream_id,
                    event_type=EventType.STREAM_OPENED,
                    data={"stream_id": stream_id, "session_id": session_id, "direction": direction},
                )
            )

        if stream_data.state in (StreamState.RESET_RECEIVED, StreamState.CLOSED):
            logger.debug("Ignoring WT data for already closed/reset stream %d", stream_id)
            return []

        if session_data is None:
            session_data = state.sessions.get(stream_data.session_id)
            if session_data is None:
                logger.error(
                    "Internal state error: Stream %d exists but session %s does not. Cannot process data.",
                    stream_id,
                    stream_data.session_id,
                )
                return effects

        if event.data:
            data_len = len(event.data)
            max_buffer_size = self._config.max_stream_read_buffer
            current_buffer_size = stream_data.read_buffer_size

            if current_buffer_size + data_len > max_buffer_size:
                logger.warning(
                    "Stream %d read buffer limit exceeded (%d + %d > %d). Sending STOP_SENDING.",
                    stream_id,
                    current_buffer_size,
                    data_len,
                    max_buffer_size,
                )
                effects.append(StopQuicStream(stream_id=stream_id, error_code=ErrorCodes.WT_FLOW_CONTROL_ERROR))
                return effects

            stream_data.bytes_received += data_len
            stream_data.read_buffer.append(event.data)
            stream_data.read_buffer_size += data_len

            session_data.peer_data_sent += data_len

        while stream_data.pending_read_requests and stream_data.read_buffer_size > 0:
            req_id, max_bytes = stream_data.pending_read_requests.popleft()
            target = self._calculate_read_amount(requested_bytes=max_bytes, buffer_size=stream_data.read_buffer_size)

            data_chunk = self._read_from_buffer(stream_data=stream_data, max_bytes=target)
            effects.append(NotifyRequestDone(request_id=req_id, result=data_chunk))

            session_data.local_data_consumed += len(data_chunk)
            credit_effect = self._check_and_send_data_credit(session_data=session_data)
            if credit_effect is not None:
                effects.append(credit_effect)

        if event.stream_ended:
            original_state = stream_data.state
            match original_state:
                case StreamState.HALF_CLOSED_LOCAL | StreamState.RESET_SENT:
                    if stream_data.read_buffer_size == 0:
                        stream_data.state = StreamState.CLOSED
                        stream_data.closed_at = get_timestamp()
                        session_data.active_streams.discard(stream_id)
                        session_data.blocked_streams.discard(stream_id)
                        effects.extend(self._process_stream_closure(session_data=session_data, stream_id=stream_id))
                        effects.append(
                            EmitStreamEvent(
                                stream_id=stream_id, event_type=EventType.STREAM_CLOSED, data={"stream_id": stream_id}
                            )
                        )
                    else:
                        stream_data.state = StreamState.HALF_CLOSED_REMOTE
                        logger.debug("Stream %d rx closed, data pending read. Moving to HALF_CLOSED_REMOTE", stream_id)
                case StreamState.OPEN:
                    stream_data.state = StreamState.HALF_CLOSED_REMOTE
                    logger.debug("Stream %d receive side closed by peer (WT data)", stream_id)

            while stream_data.pending_read_requests:
                req_id, _ = stream_data.pending_read_requests.popleft()
                effects.append(NotifyRequestDone(request_id=req_id, result=b""))

        return effects

    def _calculate_read_amount(self, *, requested_bytes: int, buffer_size: int) -> int:
        """Calculate the amount of data to read based on request and buffer size."""
        if requested_bytes <= 0:
            return buffer_size
        return min(requested_bytes, buffer_size)

    def _check_and_send_data_credit(self, *, session_data: SessionStateData, force_send: bool = False) -> Effect | None:
        """Check session data credit and send a MAX_DATA capsule if needed."""
        if session_data.state in (SessionState.CLOSED, SessionState.DRAINING):
            return None

        new_limit = calculate_new_data_limit(
            current_limit=session_data.local_max_data,
            consumed=session_data.local_data_consumed,
            window_size=self._config.flow_control_window_size,
            auto_scale=self._config.flow_control_window_auto_scale,
            force_update=force_send,
        )

        if new_limit is not None:
            logger.debug(
                "Session %s data credit auto-increment: consumed=%d limit=%d new_limit=%d",
                session_data.session_id,
                session_data.local_data_consumed,
                session_data.local_max_data,
                new_limit,
            )

            session_data.local_max_data = new_limit

            buf = QuicBuffer(capacity=8)
            buf.push_uint_var(new_limit)

            return SendH3Capsule(
                stream_id=session_data.session_id,
                capsule_type=constants.WT_MAX_DATA_TYPE,
                capsule_data=buf.data,
                end_stream=False,
            )
        return None

    def _check_and_send_stream_credit(
        self, *, session_data: SessionStateData, is_unidirectional: bool, force_send: bool = False
    ) -> Effect | None:
        """Check stream credit and send a MAX_STREAMS capsule if needed."""
        if session_data.state in (SessionState.CLOSED, SessionState.DRAINING):
            return None

        if is_unidirectional:
            current_limit = session_data.local_max_streams_uni
            closed_count = session_data.peer_streams_uni_closed
            target_window = self._config.initial_max_streams_uni
            capsule_type = constants.WT_MAX_STREAMS_UNI_TYPE
        else:
            current_limit = session_data.local_max_streams_bidi
            closed_count = session_data.peer_streams_bidi_closed
            target_window = self._config.initial_max_streams_bidi
            capsule_type = constants.WT_MAX_STREAMS_BIDI_TYPE

        new_limit = calculate_new_stream_limit(
            current_limit=current_limit,
            closed_count=closed_count,
            initial_window=target_window,
            auto_scale=self._config.flow_control_window_auto_scale,
            force_update=force_send,
        )

        if new_limit is not None:
            logger.debug(
                "Session %s stream credit auto-increment: type=%s closed=%d limit=%d new_limit=%d",
                session_data.session_id,
                "uni" if is_unidirectional else "bidi",
                closed_count,
                current_limit,
                new_limit,
            )

            if is_unidirectional:
                session_data.local_max_streams_uni = new_limit
            else:
                session_data.local_max_streams_bidi = new_limit

            buf = QuicBuffer(capacity=8)
            buf.push_uint_var(new_limit)

            return SendH3Capsule(
                stream_id=session_data.session_id, capsule_type=capsule_type, capsule_data=buf.data, end_stream=False
            )
        return None

    def _process_stream_closure(self, *, session_data: SessionStateData, stream_id: int) -> list[Effect]:
        """Update stream closure counters and trigger credit checks."""
        effects: list[Effect] = []

        is_peer_initiated = is_peer_initiated_stream(stream_id=stream_id, is_client=self._is_client)

        if is_peer_initiated:
            is_uni = is_unidirectional_stream(stream_id=stream_id)
            credit_effect: Effect | None = None

            if is_uni:
                session_data.peer_streams_uni_closed += 1
                credit_effect = self._check_and_send_stream_credit(session_data=session_data, is_unidirectional=True)
            else:
                session_data.peer_streams_bidi_closed += 1
                credit_effect = self._check_and_send_stream_credit(session_data=session_data, is_unidirectional=False)

            if credit_effect is not None:
                effects.append(credit_effect)

        return effects

    def _read_from_buffer(self, *, stream_data: StreamStateData, max_bytes: int) -> bytes:
        """Read up to max_bytes from the stream's read buffer."""
        if not stream_data.read_buffer:
            return b""

        head_chunk = stream_data.read_buffer[0]
        head_len = len(head_chunk)

        if head_len >= max_bytes:
            if head_len == max_bytes or head_len <= OPTIMIZED_READ_SLICE_THRESHOLD:
                stream_data.read_buffer.popleft()
                stream_data.read_buffer_size -= max_bytes

                if head_len == max_bytes:
                    return bytes(head_chunk)

                result = bytes(head_chunk[:max_bytes])
                remainder = head_chunk[max_bytes:]
                stream_data.read_buffer.appendleft(remainder)
                return result

        chunks: list[Buffer] = []
        bytes_collected = 0

        while stream_data.read_buffer and bytes_collected < max_bytes:
            chunk = stream_data.read_buffer.popleft()
            chunk_len = len(chunk)
            needed = max_bytes - bytes_collected

            if chunk_len <= needed:
                chunks.append(chunk)
                bytes_collected += chunk_len
                stream_data.read_buffer_size -= chunk_len
            else:
                part = chunk[:needed]
                remainder = chunk[needed:]
                chunks.append(part)
                bytes_collected += needed
                stream_data.read_buffer_size -= needed
                stream_data.read_buffer.appendleft(remainder)
                break

        return b"".join(chunks)



"""An async-native WebTransport stack for Python."""

from .client import WebTransportClient
from .config import ClientConfig, ServerConfig
from .constants import ErrorCodes
from .events import Event
from .exceptions import (
    ClientError,
    ConfigurationError,
    ConnectionError,
    DatagramError,
    ProtocolError,
    ServerError,
    SessionError,
    StreamError,
    TimeoutError,
    WebTransportError,
)
from .messaging import StructuredDatagramTransport, StructuredStream
from .server import ServerApp
from .session import WebTransportSession
from .stream import WebTransportReceiveStream, WebTransportSendStream, WebTransportStream
from .types import URL, Address, Headers
from .version import __version__

__all__: list[str] = [
    "Address",
    "ClientConfig",
    "ClientError",
    "ConfigurationError",
    "ConnectionError",
    "DatagramError",
    "ErrorCodes",
    "Event",
    "Headers",
    "ProtocolError",
    "ServerApp",
    "ServerConfig",
    "ServerError",
    "SessionError",
    "StreamError",
    "StructuredDatagramTransport",
    "StructuredStream",
    "TimeoutError",
    "URL",
    "WebTransportClient",
    "WebTransportError",
    "WebTransportReceiveStream",
    "WebTransportSendStream",
    "WebTransportSession",
    "WebTransportStream",
    "__version__",
]



"""Serializer implementation using the MsgPack format."""

from __future__ import annotations

import uuid
from dataclasses import asdict, is_dataclass
from datetime import datetime
from enum import Enum
from typing import Any, cast

from pywebtransport.exceptions import ConfigurationError, SerializationError
from pywebtransport.serializer._base import BaseDataclassSerializer
from pywebtransport.types import Buffer, Serializer

try:
    import msgpack
except ImportError:
    msgpack = None


__all__: list[str] = ["MsgPackSerializer"]


class MsgPackSerializer(BaseDataclassSerializer, Serializer):
    """Serializer for encoding and decoding using the MsgPack format."""

    def __init__(
        self, *, pack_kwargs: dict[str, Any] | None = None, unpack_kwargs: dict[str, Any] | None = None
    ) -> None:
        """Initialize the MsgPack serializer."""
        if msgpack is None:
            raise ConfigurationError(
                message="The 'msgpack' library is required for MsgPackSerializer.",
                config_key="dependency.msgpack",
                details={"installation_guide": "Please install it with: pip install pywebtransport[msgpack]"},
            )

        self._pack_kwargs = pack_kwargs.copy() if pack_kwargs is not None else {}
        self._unpack_kwargs = unpack_kwargs.copy() if unpack_kwargs is not None else {}
        self._user_default = self._pack_kwargs.pop("default", None)

    def deserialize(self, *, data: Buffer, obj_type: Any = None) -> Any:
        """Deserialize a MsgPack byte string into a Python object."""
        try:
            unpack_kwargs = {"raw": False, **self._unpack_kwargs}
            decoded_obj = msgpack.unpackb(packed=data, **unpack_kwargs)

            if obj_type is None:
                return decoded_obj
            return self.convert_to_type(data=decoded_obj, target_type=obj_type)
        except (msgpack.UnpackException, TypeError, ValueError) as e:
            raise SerializationError(
                message="Data is not valid MsgPack or cannot be unpacked.", original_exception=e
            ) from e

    def serialize(self, *, obj: Any) -> bytes:
        """Serialize a Python object into a MsgPack byte string."""
        try:
            return cast(bytes, msgpack.packb(o=obj, default=self._default_handler, **self._pack_kwargs))
        except TypeError as e:
            raise SerializationError(
                message=f"Object of type {type(obj).__name__} is not MsgPack serializable.", original_exception=e
            ) from e

    def _default_handler(self, o: Any) -> Any:
        """Handle types not natively supported by MsgPack."""
        match o:
            case uuid.UUID():
                return str(o)
            case Enum():
                return o.value
            case set() | frozenset():
                return list(o)
            case datetime():
                return o.isoformat()
            case _ if is_dataclass(o) and not isinstance(o, type):
                return asdict(obj=o)
            case _:
                if self._user_default is not None:
                    return self._user_default(o)
                raise TypeError(f"Object of type {type(o).__name__} is not MsgPack serializable")



"""Pluggable serializers for structured data transmission."""

from .json import JSONSerializer
from .msgpack import MsgPackSerializer
from .protobuf import ProtobufSerializer

__all__: list[str] = ["JSONSerializer", "MsgPackSerializer", "ProtobufSerializer"]



"""Internal state dataclasses for the protocol engine."""

from __future__ import annotations

from collections import deque
from dataclasses import dataclass, field

from pywebtransport._protocol.events import ProtocolEvent
from pywebtransport.types import (
    Buffer,
    ConnectionState,
    Headers,
    RequestId,
    SessionId,
    SessionState,
    StreamDirection,
    StreamId,
    StreamState,
)

__all__: list[str] = []


@dataclass(kw_only=True, slots=True)
class StreamStateData:
    """Represent the complete state of a single WebTransport stream."""

    stream_id: StreamId
    session_id: SessionId
    direction: StreamDirection
    state: StreamState
    created_at: float

    bytes_sent: int = 0
    bytes_received: int = 0

    read_buffer: deque[Buffer] = field(default_factory=deque)
    read_buffer_size: int = 0

    pending_read_requests: deque[tuple[RequestId, int]] = field(default_factory=deque)
    write_buffer: deque[tuple[Buffer, RequestId, bool]] = field(default_factory=deque)
    write_buffer_size: int = 0

    close_code: int | None = None
    close_reason: str | None = None
    closed_at: float | None = None


@dataclass(kw_only=True, slots=True)
class SessionInitData:
    """Temporary storage for session configuration during creation."""

    path: str
    headers: Headers
    created_at: float


@dataclass(kw_only=True, slots=True)
class SessionStateData:
    """Represent the complete state of a single WebTransport session."""

    session_id: SessionId
    state: SessionState
    path: str
    headers: Headers
    created_at: float

    local_max_data: int
    local_data_sent: int = 0
    local_data_consumed: int = 0
    peer_max_data: int
    peer_data_sent: int = 0

    local_max_streams_bidi: int
    local_streams_bidi_opened: int = 0
    peer_max_streams_bidi: int
    peer_streams_bidi_opened: int = 0
    peer_streams_bidi_closed: int = 0

    local_max_streams_uni: int
    local_streams_uni_opened: int = 0
    peer_max_streams_uni: int
    peer_streams_uni_opened: int = 0
    peer_streams_uni_closed: int = 0

    pending_bidi_stream_requests: deque[RequestId] = field(default_factory=deque)
    pending_uni_stream_requests: deque[RequestId] = field(default_factory=deque)

    datagrams_sent: int = 0
    datagram_bytes_sent: int = 0
    datagrams_received: int = 0
    datagram_bytes_received: int = 0

    active_streams: set[StreamId] = field(default_factory=set)
    blocked_streams: set[StreamId] = field(default_factory=set)

    close_code: int | None = None
    close_reason: str | None = None
    closed_at: float | None = None
    ready_at: float | None = None


@dataclass(kw_only=True, slots=True)
class ProtocolState:
    """Represent the single source of truth for an entire connection."""

    is_client: bool
    connection_state: ConnectionState
    max_datagram_size: int
    remote_max_datagram_frame_size: int = 0

    handshake_complete: bool = False
    peer_settings_received: bool = False
    local_goaway_sent: bool = False

    sessions: dict[SessionId, SessionStateData] = field(default_factory=dict)
    streams: dict[StreamId, StreamStateData] = field(default_factory=dict)

    pending_requests: dict[StreamId, RequestId] = field(default_factory=dict)
    pending_session_configs: dict[RequestId, SessionInitData] = field(default_factory=dict)

    early_event_buffer: dict[StreamId, list[tuple[float, ProtocolEvent]]] = field(default_factory=dict)
    early_event_count: int = 0

    peer_initial_max_data: int = 0
    peer_initial_max_streams_bidi: int = 0
    peer_initial_max_streams_uni: int = 0

    connected_at: float | None = None
    closed_at: float | None = None



"""Internal protocol engine for driving the WebTransport state machine."""

from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING, Any

from aioquic.buffer import encode_uint_var

from pywebtransport import constants
from pywebtransport._protocol.connection_processor import ConnectionProcessor
from pywebtransport._protocol.events import (
    CapsuleReceived,
    ConnectionClose,
    ConnectStreamClosed,
    DatagramReceived,
    Effect,
    EmitConnectionEvent,
    GoawayReceived,
    HeadersReceived,
    InternalBindH3Session,
    InternalBindQuicStream,
    InternalCleanupEarlyEvents,
    InternalCleanupResources,
    InternalFailH3Session,
    InternalFailQuicStream,
    InternalReturnStreamData,
    LogH3Frame,
    NotifyRequestFailed,
    ProtocolEvent,
    RescheduleQuicTimer,
    SendQuicData,
    SendQuicDatagram,
    SettingsReceived,
    TransportConnectionTerminated,
    TransportDatagramFrameReceived,
    TransportHandshakeCompleted,
    TransportQuicParametersReceived,
    TransportQuicTimerFired,
    TransportStreamDataReceived,
    TransportStreamReset,
    TriggerQuicTimer,
    UserAcceptSession,
    UserCloseSession,
    UserConnectionGracefulClose,
    UserCreateSession,
    UserCreateStream,
    UserEvent,
    UserGetConnectionDiagnostics,
    UserGetSessionDiagnostics,
    UserGetStreamDiagnostics,
    UserGrantDataCredit,
    UserGrantStreamsCredit,
    UserRejectSession,
    UserResetStream,
    UserSendDatagram,
    UserSendStreamData,
    UserStopStream,
    UserStreamRead,
    WebTransportStreamDataReceived,
)
from pywebtransport._protocol.h3_engine import WebTransportH3Engine
from pywebtransport._protocol.session_processor import SessionProcessor
from pywebtransport._protocol.state import ProtocolState
from pywebtransport._protocol.stream_processor import StreamProcessor
from pywebtransport.constants import ErrorCodes
from pywebtransport.exceptions import ConnectionError
from pywebtransport.types import Buffer, ConnectionId, ConnectionState, EventType, Headers, StreamId
from pywebtransport.utils import get_logger, get_timestamp, merge_headers

if TYPE_CHECKING:
    from pywebtransport.config import ClientConfig, ServerConfig


__all__: list[str] = []

logger = get_logger(name=__name__)


class WebTransportEngine:
    """Orchestrates the unified protocol state machine."""

    def __init__(self, *, connection_id: ConnectionId, config: ClientConfig | ServerConfig, is_client: bool) -> None:
        """Initialize the WebTransport engine."""
        self._connection_id = connection_id
        self._is_client = is_client
        self._config = config

        self._state = ProtocolState(
            is_client=is_client, connection_state=ConnectionState.IDLE, max_datagram_size=config.max_datagram_size
        )

        self._connection_processor = ConnectionProcessor(
            is_client=is_client, config=config, connection_id=connection_id
        )
        self._session_processor = SessionProcessor(is_client=is_client, config=config)
        self._stream_processor = StreamProcessor(is_client=is_client, config=config)
        self._h3_engine = WebTransportH3Engine(is_client=is_client, config=config)

        self._pending_user_actions: deque[UserEvent[Any]] = deque()

    def cleanup_stream(self, *, stream_id: StreamId) -> None:
        """Clean up H3 state for a closed stream."""
        self._h3_engine.cleanup_stream(stream_id=stream_id)

    def encode_capsule(
        self, *, stream_id: StreamId, capsule_type: int, capsule_data: bytes, end_stream: bool = False
    ) -> list[Effect]:
        """Encode a capsule and return effects to send it."""
        data = self._h3_engine.encode_capsule(stream_id=stream_id, capsule_type=capsule_type, capsule_data=capsule_data)
        return [SendQuicData(stream_id=stream_id, data=data, end_stream=end_stream)]

    def encode_datagram(self, *, stream_id: StreamId, data: Buffer | list[Buffer]) -> list[Effect]:
        """Encode a datagram and return effects to send it."""
        payload = self._h3_engine.encode_datagram(stream_id=stream_id, data=data)
        return [SendQuicDatagram(data=payload)]

    def encode_goaway(self) -> list[Effect]:
        """Encode a GOAWAY frame and return effects to send it."""
        control_id = self._h3_engine._local_control_stream_id
        if control_id is None:
            return []
        data = self._h3_engine.encode_goaway_frame()
        return [SendQuicData(stream_id=control_id, data=data, end_stream=False)]

    def encode_headers(self, *, stream_id: StreamId, status: int, end_stream: bool = False) -> list[Effect]:
        """Encode headers and return effects to send them."""
        headers: Headers = {":status": str(status)}
        return self._h3_engine.encode_headers(stream_id=stream_id, headers=headers, end_stream=end_stream)

    def encode_session_request(
        self, *, stream_id: StreamId, path: str, authority: str, headers: Headers
    ) -> list[Effect]:
        """Encode a WebTransport session establishment request (CONNECT)."""
        initial_headers: Headers = {
            ":method": "CONNECT",
            ":scheme": "https",
            ":authority": authority,
            ":path": path,
            ":protocol": "webtransport",
        }
        final_headers = merge_headers(base=initial_headers, update=headers)
        return self._h3_engine.encode_headers(stream_id=stream_id, headers=final_headers, end_stream=False)

    def encode_stream_creation(
        self, *, stream_id: StreamId, control_stream_id: StreamId, is_unidirectional: bool
    ) -> list[Effect]:
        """Encode the preamble for a new WebTransport stream."""
        return self._h3_engine.encode_webtransport_stream_creation(
            stream_id=stream_id, control_stream_id=control_stream_id, is_unidirectional=is_unidirectional
        )

    def handle_event(self, *, event: ProtocolEvent) -> list[Effect]:
        """Process a single event and return resulting effects."""
        all_effects: list[Effect] = []
        events_to_process: deque[ProtocolEvent] = deque([event])

        while events_to_process:
            current_event = events_to_process.popleft()
            new_effects: list[Effect] = []
            re_queue_pending_actions = False

            match current_event:
                case InternalBindH3Session() as ibhs_ev:
                    new_effects.extend(
                        self._connection_processor.handle_internal_bind_h3_session(event=ibhs_ev, state=self._state)
                    )

                case InternalBindQuicStream() as ibqs_ev:
                    new_effects.extend(
                        self._stream_processor.handle_internal_bind_quic_stream(event=ibqs_ev, state=self._state)
                    )

                case InternalCleanupEarlyEvents() as icee_ev:
                    new_effects.extend(
                        self._connection_processor.handle_cleanup_early_events(event=icee_ev, state=self._state)
                    )

                case InternalCleanupResources() as icr_ev:
                    new_effects.extend(
                        self._connection_processor.handle_cleanup_resources(event=icr_ev, state=self._state)
                    )

                case InternalFailH3Session() as ifhs_ev:
                    new_effects.extend(
                        self._connection_processor.handle_internal_fail_h3_session(event=ifhs_ev, state=self._state)
                    )

                case InternalFailQuicStream() as ifqs_ev:
                    new_effects.extend(
                        self._stream_processor.handle_internal_fail_quic_stream(event=ifqs_ev, state=self._state)
                    )

                case InternalReturnStreamData() as irsd_ev:
                    new_effects.extend(
                        self._stream_processor.handle_return_stream_data(event=irsd_ev, state=self._state)
                    )

                case TransportConnectionTerminated() as tct_ev:
                    new_effects.extend(
                        self._connection_processor.handle_connection_terminated(event=tct_ev, state=self._state)
                    )
                    error = ConnectionError(
                        message=f"Connection terminated before ready: {tct_ev.reason_phrase}",
                        error_code=tct_ev.error_code,
                    )
                    new_effects.extend(self._fail_pending_user_actions(exception=error))

                case TransportDatagramFrameReceived() | TransportStreamDataReceived() as tev:
                    was_settings_received = self._h3_engine._settings_received

                    h3_events, h3_effects = self._h3_engine.handle_transport_event(event=tev, state=self._state)
                    new_effects.extend(h3_effects)
                    events_to_process.extendleft(reversed(h3_events))

                    if self._is_client and not was_settings_received and self._h3_engine._settings_received:
                        logger.debug("Client received peer H3 SETTINGS.")
                        self._state.peer_settings_received = True
                        (readiness_effects, is_ready) = self._check_client_connection_ready()
                        new_effects.extend(readiness_effects)
                        if is_ready:
                            re_queue_pending_actions = True

                case TransportHandshakeCompleted():
                    if self._state.connection_state == ConnectionState.IDLE:
                        logger.debug("State transition: IDLE -> CONNECTING")
                        self._state.connection_state = ConnectionState.CONNECTING

                    if self._state.connection_state == ConnectionState.CONNECTING:
                        logger.debug("TransportHandshakeCompleted received.")
                        self._state.handshake_complete = True

                        if self._is_client:
                            (readiness_effects, is_ready) = self._check_client_connection_ready()
                            new_effects.extend(readiness_effects)
                            if is_ready:
                                re_queue_pending_actions = True
                        else:
                            self._state.connection_state = ConnectionState.CONNECTED
                            self._state.connected_at = get_timestamp()
                            new_effects.append(
                                EmitConnectionEvent(
                                    event_type=EventType.CONNECTION_ESTABLISHED,
                                    data={"connection_id": self._connection_id},
                                )
                            )
                    else:
                        logger.warning(
                            "Received TransportHandshakeCompleted in unexpected state: %s", self._state.connection_state
                        )

                case TransportQuicParametersReceived() as tqpr_ev:
                    new_effects.extend(
                        self._connection_processor.handle_transport_parameters_received(
                            event=tqpr_ev, state=self._state
                        )
                    )

                case TransportQuicTimerFired():
                    new_effects.extend([TriggerQuicTimer(), RescheduleQuicTimer()])

                case TransportStreamReset() as tsr_ev:
                    new_effects.extend(
                        self._stream_processor.handle_transport_stream_reset(event=tsr_ev, state=self._state)
                    )

                case CapsuleReceived() as cr_ev:
                    new_effects.extend(self._session_processor.handle_capsule_received(event=cr_ev, state=self._state))

                case ConnectStreamClosed() as csc_ev:
                    new_effects.extend(
                        self._session_processor.handle_connect_stream_closed(event=csc_ev, state=self._state)
                    )

                case DatagramReceived() as dr_ev:
                    new_effects.extend(self._session_processor.handle_datagram_received(event=dr_ev, state=self._state))

                case GoawayReceived() as gr_ev:
                    new_effects.extend(
                        self._connection_processor.handle_goaway_received(event=gr_ev, state=self._state)
                    )

                case HeadersReceived() as hr_ev:
                    new_effects.extend(
                        self._connection_processor.handle_headers_received(event=hr_ev, state=self._state)
                    )

                case SettingsReceived() as sr_ev:
                    logger.debug("Processing H3 SETTINGS frame.")
                    self._state.peer_initial_max_data = sr_ev.settings.get(constants.SETTINGS_WT_INITIAL_MAX_DATA, 0)
                    self._state.peer_initial_max_streams_bidi = sr_ev.settings.get(
                        constants.SETTINGS_WT_INITIAL_MAX_STREAMS_BIDI, 0
                    )
                    self._state.peer_initial_max_streams_uni = sr_ev.settings.get(
                        constants.SETTINGS_WT_INITIAL_MAX_STREAMS_UNI, 0
                    )

                case WebTransportStreamDataReceived() as wtsdr_ev:
                    new_effects.extend(
                        self._stream_processor.handle_webtransport_stream_data(event=wtsdr_ev, state=self._state)
                    )

                case UserAcceptSession() as uas_ev:
                    new_effects.extend(self._session_processor.handle_accept_session(event=uas_ev, state=self._state))

                case UserCloseSession() as ucs_ev:
                    new_effects.extend(self._session_processor.handle_close_session(event=ucs_ev, state=self._state))

                case UserConnectionGracefulClose() as ugcc_ev:
                    new_effects.extend(
                        self._connection_processor.handle_graceful_close(event=ugcc_ev, state=self._state)
                    )

                case UserCreateSession() as ucs_ev:
                    if self._is_client and self._state.connection_state in (
                        ConnectionState.IDLE,
                        ConnectionState.CONNECTING,
                    ):
                        logger.debug("Client not fully connected, buffering UserCreateSession.")
                        self._pending_user_actions.append(ucs_ev)
                    else:
                        new_effects.extend(
                            self._connection_processor.handle_create_session(event=ucs_ev, state=self._state)
                        )

                case UserCreateStream() as ucs_ev:
                    if self._is_client and self._state.connection_state in (
                        ConnectionState.IDLE,
                        ConnectionState.CONNECTING,
                    ):
                        logger.debug("Client not fully connected, buffering UserCreateStream.")
                        self._pending_user_actions.append(ucs_ev)
                    else:
                        new_effects.extend(
                            self._session_processor.handle_create_stream(event=ucs_ev, state=self._state)
                        )

                case UserGetConnectionDiagnostics() as ugcd_ev:
                    new_effects.extend(
                        self._connection_processor.handle_get_connection_diagnostics(event=ugcd_ev, state=self._state)
                    )

                case UserGetSessionDiagnostics() as ugsd_ev:
                    new_effects.extend(
                        self._session_processor.handle_get_session_diagnostics(event=ugsd_ev, state=self._state)
                    )

                case UserGetStreamDiagnostics() as ugstd_ev:
                    new_effects.extend(
                        self._stream_processor.handle_get_stream_diagnostics(event=ugstd_ev, state=self._state)
                    )

                case UserGrantDataCredit() as ugdc_ev:
                    new_effects.extend(
                        self._session_processor.handle_grant_data_credit(event=ugdc_ev, state=self._state)
                    )

                case UserGrantStreamsCredit() as ugsc_ev:
                    new_effects.extend(
                        self._session_processor.handle_grant_streams_credit(event=ugsc_ev, state=self._state)
                    )

                case UserRejectSession() as urs_ev:
                    new_effects.extend(self._session_processor.handle_reject_session(event=urs_ev, state=self._state))

                case UserResetStream() as urs_ev:
                    new_effects.extend(self._stream_processor.handle_reset_stream(event=urs_ev, state=self._state))

                case UserSendDatagram() as usd_ev:
                    new_effects.extend(self._session_processor.handle_send_datagram(event=usd_ev, state=self._state))

                case UserSendStreamData() as ussd_ev:
                    new_effects.extend(self._stream_processor.handle_send_stream_data(event=ussd_ev, state=self._state))

                case UserStopStream() as uss_ev:
                    new_effects.extend(self._stream_processor.handle_stop_stream(event=uss_ev, state=self._state))

                case UserStreamRead() as usr_ev:
                    new_effects.extend(self._stream_processor.handle_stream_read(event=usr_ev, state=self._state))

                case ConnectionClose() as cc_ev:
                    new_effects.extend(
                        self._connection_processor.handle_connection_close(event=cc_ev, state=self._state)
                    )
                    error = ConnectionError(message="Connection closed by application", error_code=ErrorCodes.NO_ERROR)
                    new_effects.extend(self._fail_pending_user_actions(exception=error))

                case _:
                    logger.warning("Unhandled event type in engine's handle_event: %s", type(current_event))

            all_effects.extend(new_effects)

            if re_queue_pending_actions:
                if self._pending_user_actions:
                    logger.debug(
                        "Connection is ready, re-queueing %d pending user actions.", len(self._pending_user_actions)
                    )
                    events_to_process.extendleft(reversed(self._pending_user_actions))
                    self._pending_user_actions.clear()

        all_effects.append(RescheduleQuicTimer())

        return all_effects

    def initialize_h3_transport(
        self, *, control_id: StreamId, encoder_id: StreamId, decoder_id: StreamId
    ) -> list[Effect]:
        """Initialize HTTP/3 unidirectional streams and settings."""
        self._h3_engine.set_local_stream_ids(
            control_stream_id=control_id, encoder_stream_id=encoder_id, decoder_stream_id=decoder_id
        )

        settings_bytes = self._h3_engine.initialize_connection()

        effects: list[Effect] = [
            SendQuicData(
                stream_id=control_id,
                data=encode_uint_var(constants.H3_STREAM_TYPE_CONTROL) + settings_bytes,
                end_stream=False,
            ),
            SendQuicData(
                stream_id=encoder_id, data=encode_uint_var(constants.H3_STREAM_TYPE_QPACK_ENCODER), end_stream=False
            ),
            SendQuicData(
                stream_id=decoder_id, data=encode_uint_var(constants.H3_STREAM_TYPE_QPACK_DECODER), end_stream=False
            ),
            LogH3Frame(category="http", event="stream_type_set", data={"new": "control", "stream_id": control_id}),
            LogH3Frame(
                category="http", event="stream_type_set", data={"new": "qpack_encoder", "stream_id": encoder_id}
            ),
            LogH3Frame(
                category="http", event="stream_type_set", data={"new": "qpack_decoder", "stream_id": decoder_id}
            ),
        ]
        return effects

    def _check_client_connection_ready(self) -> tuple[list[Effect], bool]:
        """Check if the client connection is fully ready (QUIC + H3)."""
        if (
            self._state.connection_state == ConnectionState.CONNECTING
            and self._state.handshake_complete
            and self._state.peer_settings_received
        ):
            logger.debug("Client connection fully ready (QUIC + H3 SETTINGS).")
            self._state.connection_state = ConnectionState.CONNECTED
            self._state.connected_at = get_timestamp()
            effects: list[Effect] = [
                EmitConnectionEvent(
                    event_type=EventType.CONNECTION_ESTABLISHED, data={"connection_id": self._connection_id}
                )
            ]
            return effects, True

        return [], False

    def _fail_pending_user_actions(self, *, exception: Exception) -> list[Effect]:
        """Fail all pending user actions with the given exception."""
        effects: list[Effect] = []
        while self._pending_user_actions:
            action = self._pending_user_actions.popleft()
            effects.append(NotifyRequestFailed(request_id=action.request_id, exception=exception))
        return effects



"""Shared, general-purpose utilities."""

from __future__ import annotations

import logging
import os
import time
from datetime import datetime, timedelta, timezone
from pathlib import Path
from types import TracebackType
from typing import Self

from pywebtransport.types import Buffer, Headers

__all__: list[str] = [
    "Timer",
    "ensure_buffer",
    "format_duration",
    "generate_self_signed_cert",
    "get_header",
    "get_header_as_str",
    "get_logger",
    "get_timestamp",
    "merge_headers",
]

_timer_logger = logging.getLogger(name="timer")


class Timer:
    """A simple context manager for performance measurement using a monotonic clock."""

    def __init__(self, *, name: str = "timer") -> None:
        """Initialize the timer."""
        self.name = name
        self.start_time: float | None = None
        self.end_time: float | None = None

    @property
    def elapsed(self) -> float:
        """Get the elapsed time in seconds."""
        if self.start_time is None:
            return 0.0

        end = self.end_time if self.end_time is not None else time.perf_counter()
        return end - self.start_time

    def __enter__(self) -> Self:
        """Start the timer upon entering the context."""
        self.start()
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Stop the timer and log the duration upon exiting the context."""
        elapsed = self.stop()
        _timer_logger.debug("%s took %s", self.name, format_duration(seconds=elapsed))

    def start(self) -> None:
        """Start the timer."""
        self.start_time = time.perf_counter()
        self.end_time = None

    def stop(self) -> float:
        """Stop the timer and return the elapsed time."""
        if self.start_time is None:
            raise RuntimeError("Timer not started")

        self.end_time = time.perf_counter()
        return self.elapsed


def ensure_buffer(*, data: Buffer | str, encoding: str = "utf-8") -> Buffer:
    """Ensure that the given data is in a buffer-compatible format."""
    match data:
        case str():
            return data.encode(encoding=encoding)
        case bytes() | bytearray() | memoryview():
            return data
        case _:
            raise TypeError(f"Expected str or Buffer, got {type(data).__name__}")


def format_duration(*, seconds: float) -> str:
    """Format a duration in seconds into a human-readable string."""
    if seconds < 1e-6:
        return f"{seconds * 1e9:.0f}ns"
    if seconds < 1e-3:
        return f"{seconds * 1e6:.1f}µs"
    if seconds < 1:
        return f"{seconds * 1000:.1f}ms"
    if seconds < 60:
        return f"{seconds:.1f}s"
    if seconds < 3600:
        minutes = int(seconds // 60)
        secs = seconds % 60
        return f"{minutes}m{secs:.1f}s"

    hours = int(seconds // 3600)
    minutes = int((seconds % 3600) // 60)
    secs = seconds % 60
    return f"{hours}h{minutes}m{secs:.1f}s"


def generate_self_signed_cert(
    *, hostname: str, output_dir: str = ".", days_valid: int = 365, key_size: int = 2048
) -> tuple[str, str]:
    """Generate a self-signed certificate and key for testing purposes."""
    from cryptography import x509
    from cryptography.hazmat.primitives import hashes, serialization
    from cryptography.hazmat.primitives.asymmetric import rsa
    from cryptography.x509.oid import NameOID

    private_key = rsa.generate_private_key(public_exponent=65537, key_size=key_size)
    subject = issuer = x509.Name(
        attributes=[
            x509.NameAttribute(oid=NameOID.COUNTRY_NAME, value="US"),
            x509.NameAttribute(oid=NameOID.STATE_OR_PROVINCE_NAME, value="CA"),
            x509.NameAttribute(oid=NameOID.LOCALITY_NAME, value="San Francisco"),
            x509.NameAttribute(oid=NameOID.ORGANIZATION_NAME, value="PyWebTransport"),
            x509.NameAttribute(oid=NameOID.COMMON_NAME, value=hostname),
        ]
    )

    cert = (
        x509.CertificateBuilder()
        .subject_name(name=subject)
        .issuer_name(name=issuer)
        .public_key(key=private_key.public_key())
        .serial_number(number=x509.random_serial_number())
        .not_valid_before(time=datetime.now(tz=timezone.utc))
        .not_valid_after(time=datetime.now(tz=timezone.utc) + timedelta(days=days_valid))
        .add_extension(extval=x509.SubjectAlternativeName([x509.DNSName(value=hostname)]), critical=False)
        .sign(private_key=private_key, algorithm=hashes.SHA256())
    )

    output_path = Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)
    cert_file = output_path / f"{hostname}.crt"
    key_file = output_path / f"{hostname}.key"

    with open(file=key_file, mode="wb") as f:
        f.write(
            private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )
    os.chmod(path=key_file, mode=0o600)

    with open(file=cert_file, mode="wb") as f:
        f.write(cert.public_bytes(encoding=serialization.Encoding.PEM))

    return (str(cert_file), str(key_file))


def get_header(*, headers: Headers, key: str, default: str | bytes | None = None) -> str | bytes | None:
    """
    Get a header value case-insensitively from a dict or list.

    .. note::
        Performance Optimization: When ``headers`` is a dict, this function assumes
        keys are already lowercased (per HTTP/3 spec) to ensure O(1) lookup.
    """
    target_key = key.lower()
    target_key_bytes = target_key.encode("utf-8")

    if isinstance(headers, dict):
        if target_key in headers:
            return headers[target_key]
        return headers.get(target_key_bytes, default)

    for k, v in headers:
        if isinstance(k, bytes):
            if k.lower() == target_key_bytes:
                return v
        elif k.lower() == target_key:
            return v
    return default


def get_header_as_str(*, headers: Headers, key: str, default: str | None = None) -> str | None:
    """Get a header value and decode it to a string if necessary."""
    value = get_header(headers=headers, key=key)
    if value is None:
        return default

    if isinstance(value, str):
        return value

    try:
        return value.decode("utf-8")
    except UnicodeDecodeError:
        return default


def get_logger(*, name: str) -> logging.Logger:
    """Get a logger instance with a specific name."""
    return logging.getLogger(name=name)


def get_timestamp() -> float:
    """Get the current monotonic timestamp."""
    return time.perf_counter()


def merge_headers(*, base: Headers, update: Headers | None) -> Headers:
    """Merge two sets of headers, preserving list format if present."""
    if update is None:
        if isinstance(base, dict):
            return base.copy()
        return list(base)

    if isinstance(base, dict) and isinstance(update, dict):
        new_headers = base.copy()
        new_headers.update(update)
        return new_headers

    base_list = list(base.items()) if isinstance(base, dict) else list(base)
    update_list = list(update.items()) if isinstance(update, dict) else list(update)
    return base_list + update_list



"""Core server implementation for accepting WebTransport connections."""

from __future__ import annotations

import asyncio
from asyncio import BaseTransport, DatagramTransport
from collections import Counter
from dataclasses import asdict, dataclass
from pathlib import Path
from types import TracebackType
from typing import Any, Self, cast

from aioquic.asyncio.server import QuicServer

from pywebtransport._adapter.server import WebTransportServerProtocol, create_server
from pywebtransport.config import ServerConfig
from pywebtransport.connection import WebTransportConnection
from pywebtransport.events import Event, EventEmitter
from pywebtransport.exceptions import ServerError
from pywebtransport.manager.connection import ConnectionManager
from pywebtransport.manager.session import SessionManager
from pywebtransport.types import Address, ConnectionState, EventType, SessionState
from pywebtransport.utils import get_logger, get_timestamp

__all__ = ["ServerDiagnostics", "ServerStats", "WebTransportServer"]

logger = get_logger(name=__name__)


@dataclass(frozen=True, kw_only=True)
class ServerDiagnostics:
    """A structured, immutable snapshot of a server's health."""

    is_serving: bool
    stats: ServerStats
    connection_states: dict[ConnectionState, int]
    max_connections: int
    session_states: dict[SessionState, int]
    certfile_path: str
    cert_file_exists: bool
    keyfile_path: str
    key_file_exists: bool

    @property
    def issues(self) -> list[str]:
        """Get a list of potential issues based on the current diagnostics."""
        issues: list[str] = []
        stats_dict = self.stats.to_dict()

        if not self.is_serving:
            issues.append("Server is not currently serving.")

        total_attempts = stats_dict.get("total_connections_attempted", 0)
        success_rate = stats_dict.get("success_rate", 1.0)
        connections_rejected = stats_dict.get("connections_rejected", 0)

        if total_attempts > 20 and success_rate < 0.9:
            issues.append(f"High connection rejection rate: {connections_rejected}/{total_attempts}")

        active_connections = self.connection_states.get(ConnectionState.CONNECTED, 0)
        if self.max_connections > 0 and (active_connections / max(1, self.max_connections)) > 0.9:
            issues.append(f"High connection usage: {active_connections / self.max_connections:.1%}")

        if self.certfile_path and not self.cert_file_exists:
            issues.append(f"Certificate file not found: {self.certfile_path}")
        if self.keyfile_path and not self.key_file_exists:
            issues.append(f"Key file not found: {self.keyfile_path}")

        return issues


@dataclass(kw_only=True)
class ServerStats:
    """Represent statistics for the server."""

    start_time: float | None = None
    connections_accepted: int = 0
    connections_rejected: int = 0
    connection_errors: int = 0
    protocol_errors: int = 0

    @property
    def total_connections_attempted(self) -> int:
        """Get the total number of connections attempted."""
        return self.connections_accepted + self.connections_rejected

    @property
    def success_rate(self) -> float:
        """Get the connection success rate."""
        total = self.total_connections_attempted
        if total == 0:
            return 1.0
        return self.connections_accepted / total

    def to_dict(self) -> dict[str, Any]:
        """Convert statistics to a dictionary."""
        data = asdict(obj=self)
        data["total_connections_attempted"] = self.total_connections_attempted
        data["success_rate"] = self.success_rate
        data["uptime"] = (get_timestamp() - self.start_time) if self.start_time is not None else 0.0
        return data


class WebTransportServer(EventEmitter):
    """Manage the lifecycle and connections for the WebTransport server."""

    def __init__(self, *, config: ServerConfig | None = None) -> None:
        """Initialize the WebTransport server."""
        self._config = config if config is not None else ServerConfig()
        self._config.validate()
        super().__init__(
            max_queue_size=self._config.max_event_queue_size,
            max_listeners=self._config.max_event_listeners,
            max_history=self._config.max_event_history_size,
        )
        self._serving, self._closing = False, False
        self._server: QuicServer | None = None
        self._connection_manager = ConnectionManager(max_connections=self._config.max_connections)
        self._session_manager = SessionManager(max_sessions=self._config.max_sessions)
        self._background_tasks: set[asyncio.Task[Any]] = set()
        self._stats = ServerStats()
        self._shutdown_event = asyncio.Event()
        self._close_task: asyncio.Task[None] | None = None
        logger.info("WebTransport server initialized.")

    @property
    def config(self) -> ServerConfig:
        """Get the server's configuration object."""
        return self._config

    @property
    def connection_manager(self) -> ConnectionManager:
        """Get the server's connection manager instance."""
        return self._connection_manager

    @property
    def is_serving(self) -> bool:
        """Check if the server is currently serving."""
        return self._serving

    @property
    def local_address(self) -> Address | None:
        """Get the local address the server is bound to."""
        if self._server is not None:
            try:
                transport = self._server._transport
                if transport is not None:
                    return cast(Address | None, transport.get_extra_info("sockname"))
            except (OSError, AttributeError):
                return None
        return None

    @property
    def session_manager(self) -> SessionManager:
        """Get the server's session manager instance."""
        return self._session_manager

    async def __aenter__(self) -> Self:
        """Enter the async context for the server."""
        await self._connection_manager.__aenter__()
        await self._session_manager.__aenter__()
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Exit the async context and close the server."""
        await self.close()

    async def close(self) -> None:
        """Gracefully shut down the server and its resources."""
        if self._close_task is not None and not self._close_task.done():
            await self._close_task
            return

        if not self._serving:
            return

        self._close_task = asyncio.create_task(coro=self._close_implementation())
        await self._close_task

    async def listen(self, *, host: str | None = None, port: int | None = None) -> None:
        """Start the server and begin listening for connections."""
        if self._serving:
            raise ServerError(message="Server is already serving")

        bind_host = host if host is not None else self._config.bind_host
        bind_port = port if port is not None else self._config.bind_port

        logger.info("Starting WebTransport server on %s:%s", bind_host, bind_port)

        try:
            self._server = await create_server(
                host=bind_host, port=bind_port, config=self._config, connection_creator=self._create_connection_callback
            )
            self._serving = True
            self._stats.start_time = get_timestamp()
            logger.info("WebTransport server listening on %s", self.local_address)
        except FileNotFoundError as e:
            logger.critical("Certificate/Key file error: %s", e)
            raise ServerError(message=f"Certificate/Key file error: {e}") from e
        except Exception as e:
            logger.critical("Failed to start server: %s", e, exc_info=True)
            raise ServerError(message=f"Failed to start server: {e}") from e

    async def serve_forever(self) -> None:
        """Run the server indefinitely until interrupted."""
        if not self._serving or self._server is None:
            raise ServerError(message="Server is not listening")

        logger.info("Server is running. Press Ctrl+C to stop.")
        try:
            await self._shutdown_event.wait()
        except asyncio.CancelledError:
            logger.info("serve_forever cancelled.")
        except Exception as e:
            logger.error("Error during serve_forever wait: %s", e)
        finally:
            logger.info("serve_forever loop finished.")

    async def diagnostics(self) -> ServerDiagnostics:
        """Get a snapshot of the server's diagnostics and statistics."""
        async with asyncio.TaskGroup() as tg:
            conn_task = tg.create_task(coro=self._connection_manager.get_all_resources())
            sess_task = tg.create_task(coro=self._session_manager.get_all_resources())

        connections = conn_task.result()
        sessions = sess_task.result()
        connection_states = Counter(conn.state for conn in connections)
        session_states = Counter(sess.state for sess in sessions)

        cert_path = self.config.certfile
        key_path = self.config.keyfile

        def check_files() -> tuple[bool, bool]:
            c_exists = Path(cert_path).exists() if cert_path is not None and cert_path else False
            k_exists = Path(key_path).exists() if key_path is not None and key_path else False
            return c_exists, k_exists

        loop = asyncio.get_running_loop()
        cert_exists, key_exists = await loop.run_in_executor(None, check_files)

        return ServerDiagnostics(
            is_serving=self.is_serving,
            stats=self._stats,
            connection_states=dict(connection_states),
            max_connections=self.config.max_connections,
            session_states=dict(session_states),
            certfile_path=cert_path if cert_path is not None else "",
            cert_file_exists=cert_exists,
            keyfile_path=key_path if key_path is not None else "",
            key_file_exists=key_exists,
        )

    async def _close_implementation(self) -> None:
        """Internal implementation of server closure."""
        logger.info("Closing WebTransport server...")
        self._serving = False
        self._closing = True

        for task in self._background_tasks:
            if not task.done():
                task.cancel()

        if self._background_tasks:
            await asyncio.gather(*self._background_tasks, return_exceptions=True)

        try:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(coro=self._connection_manager.shutdown())
                tg.create_task(coro=self._session_manager.shutdown())
        except* Exception as eg:
            logger.error("Errors occurred during manager shutdown: %s", eg.exceptions, exc_info=eg)

        if self._server is not None:
            self._server.close()

        self._shutdown_event.set()

        self._closing = False
        logger.info("WebTransport server closed.")

    def _create_connection_callback(self, protocol: WebTransportServerProtocol, transport: BaseTransport) -> None:
        """Create a new WebTransportConnection from the protocol."""
        logger.debug("Creating WebTransportConnection via callback.")

        if not hasattr(transport, "sendto"):
            logger.error("Received transport without 'sendto' method: %s", type(transport).__name__)
            if not transport.is_closing():
                transport.close()
            return

        try:
            connection = WebTransportConnection.accept(
                transport=cast(DatagramTransport, transport), protocol=protocol, config=self._config
            )
            task = asyncio.create_task(coro=self._initialize_and_register_connection(connection=connection))
            self._background_tasks.add(task)
            task.add_done_callback(self._background_tasks.discard)
        except Exception as e:
            logger.error("Error creating WebTransportConnection in callback: %s", e, exc_info=True)
            if not transport.is_closing():
                transport.close()

    async def _initialize_and_register_connection(self, connection: WebTransportConnection) -> None:
        """Initialize connection engine and register with manager."""

        async def forward_session_request(event: Event) -> None:
            event_data = event.data.copy() if isinstance(event.data, dict) else {}
            event_data["connection"] = connection

            session = event_data.get("session")
            if session is not None:
                try:
                    await self._session_manager.add_session(session=session)
                except Exception as e:
                    logger.error("Failed to register session %s: %s", session.session_id, e)

            await self.emit(event_type=EventType.SESSION_REQUEST, data=event_data)

        connection.events.on(event_type=EventType.SESSION_REQUEST, handler=forward_session_request)

        try:
            await self._connection_manager.add_connection(connection=connection)
            self._stats.connections_accepted += 1
            logger.info("New connection registered: %s", connection.connection_id)
        except Exception as e:
            self._stats.connections_rejected += 1
            self._stats.connection_errors += 1
            logger.error("Failed to initialize/register new connection: %s", e, exc_info=True)
            connection.events.off(event_type=EventType.SESSION_REQUEST, handler=forward_session_request)
            if not connection.is_closed:
                await connection.close()
        else:

            async def cleanup_listener(event: Event) -> None:
                connection.events.off(event_type=EventType.SESSION_REQUEST, handler=forward_session_request)

            connection.events.once(event_type=EventType.CONNECTION_CLOSED, handler=cleanup_listener)

    def __str__(self) -> str:
        """Format a concise summary of server information for logging."""
        status = "serving" if self.is_serving else "stopped"
        address_info = self.local_address
        address_str = f"{address_info[0]}:{address_info[1]}" if address_info is not None else "unknown"
        conn_count = len(self._connection_manager)
        sess_count = len(self._session_manager)
        return (
            f"WebTransportServer(status={status}, "
            f"address={address_str}, "
            f"connections={conn_count}, "
            f"sessions={sess_count})"
        )



"""Core data types and interface protocols for the library."""

from __future__ import annotations

import asyncio
import ssl
from collections.abc import AsyncGenerator, AsyncIterator
from contextlib import AbstractAsyncContextManager as AsyncContextManager
from enum import StrEnum
from typing import Any, Protocol, runtime_checkable

__all__: list[str] = [
    "Address",
    "AsyncContextManager",
    "AsyncGenerator",
    "AsyncIterator",
    "Buffer",
    "ConnectionId",
    "ConnectionState",
    "Data",
    "ErrorCode",
    "EventData",
    "EventType",
    "Future",
    "Headers",
    "Priority",
    "RequestId",
    "SSLContext",
    "Serializer",
    "SessionId",
    "SessionProtocol",
    "SessionState",
    "StreamDirection",
    "StreamId",
    "StreamState",
    "Timeout",
    "Timestamp",
    "URL",
    "URLParts",
    "WebTransportProtocol",
    "Weight",
]


type Address = tuple[str, int]
type Buffer = bytes | bytearray | memoryview
type ConnectionId = str
type Data = bytes | bytearray | memoryview | str
type ErrorCode = int
type EventData = Any
type Future[T] = asyncio.Future[T]
type Headers = dict[str | bytes, str | bytes] | list[tuple[str | bytes, str | bytes]]
type Priority = int
type RequestId = int
type SessionId = int
type SSLContext = ssl.SSLContext
type StreamId = int
type Timeout = float | None
type Timestamp = float
type URL = str
type URLParts = tuple[str, int, str]
type Weight = int


@runtime_checkable
class Serializer(Protocol):
    """A protocol for serializing and deserializing structured data."""

    def deserialize(self, *, data: Buffer, obj_type: Any = None) -> Any:
        """Deserialize buffer into an object."""
        ...

    def serialize(self, *, obj: Any) -> bytes:
        """Serialize an object into bytes."""
        ...


@runtime_checkable
class SessionProtocol(Protocol):
    """A protocol defining the essential interface of a WebTransport session."""

    @property
    def headers(self) -> Headers:
        """Get the session headers."""
        ...

    @property
    def path(self) -> str:
        """Get the session path."""
        ...

    @property
    def remote_address(self) -> Address | None:
        """Get the remote address of the peer."""
        ...

    @property
    def session_id(self) -> SessionId:
        """Get the session ID."""
        ...

    @property
    def state(self) -> SessionState:
        """Get the current session state."""
        ...

    async def close(self, *, error_code: int = 0, reason: str | None = None) -> None:
        """Close the session."""
        ...


@runtime_checkable
class WebTransportProtocol(Protocol):
    """A protocol for the underlying WebTransport transport layer."""

    def connection_lost(self, exc: Exception | None) -> None:
        """Called when a connection is lost."""
        ...

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Called when a connection is established."""
        ...

    def datagram_received(self, data: Buffer, addr: Address) -> None:
        """Called when a datagram is received."""
        ...

    def error_received(self, exc: Exception) -> None:
        """Called when an error is received."""
        ...


class ConnectionState(StrEnum):
    """Enumeration of connection states."""

    IDLE = "idle"
    CONNECTING = "connecting"
    CONNECTED = "connected"
    CLOSING = "closing"
    CLOSED = "closed"
    FAILED = "failed"
    DRAINING = "draining"


class EventType(StrEnum):
    """Enumeration of system event types."""

    CAPSULE_RECEIVED = "capsule_received"
    CONNECTION_CLOSED = "connection_closed"
    CONNECTION_ESTABLISHED = "connection_established"
    CONNECTION_FAILED = "connection_failed"
    CONNECTION_LOST = "connection_lost"
    DATAGRAM_ERROR = "datagram_error"
    DATAGRAM_RECEIVED = "datagram_received"
    DATAGRAM_SENT = "datagram_sent"
    PROTOCOL_ERROR = "protocol_error"
    SESSION_CLOSED = "session_closed"
    SESSION_DATA_BLOCKED = "session_data_blocked"
    SESSION_DRAINING = "session_draining"
    SESSION_MAX_DATA_UPDATED = "session_max_data_updated"
    SESSION_MAX_STREAMS_BIDI_UPDATED = "session_max_streams_bidi_updated"
    SESSION_MAX_STREAMS_UNI_UPDATED = "session_max_streams_uni_updated"
    SESSION_READY = "session_ready"
    SESSION_REQUEST = "session_request"
    SESSION_STREAMS_BLOCKED = "session_streams_blocked"
    SETTINGS_RECEIVED = "settings_received"
    STREAM_CLOSED = "stream_closed"
    STREAM_DATA_RECEIVED = "stream_data_received"
    STREAM_ERROR = "stream_error"
    STREAM_OPENED = "stream_opened"
    TIMEOUT_ERROR = "timeout_error"


class SessionState(StrEnum):
    """Enumeration of WebTransport session states."""

    CONNECTING = "connecting"
    CONNECTED = "connected"
    CLOSING = "closing"
    DRAINING = "draining"
    CLOSED = "closed"


class StreamDirection(StrEnum):
    """Enumeration of stream directions."""

    BIDIRECTIONAL = "bidirectional"
    SEND_ONLY = "send_only"
    RECEIVE_ONLY = "receive_only"


class StreamState(StrEnum):
    """Enumeration of WebTransport stream states."""

    OPEN = "open"
    HALF_CLOSED_LOCAL = "half_closed_local"
    HALF_CLOSED_REMOTE = "half_closed_remote"
    CLOSED = "closed"
    RESET_SENT = "reset_sent"
    RESET_RECEIVED = "reset_received"



"""Internal specialized H3 protocol engine logic."""

from __future__ import annotations

import functools
from collections import deque
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, cast

import pylsqpack
from aioquic._buffer import Buffer as QuicBuffer
from aioquic._buffer import BufferReadError
from aioquic.buffer import UINT_VAR_MAX_SIZE, encode_uint_var

from pywebtransport import constants
from pywebtransport._protocol.events import (
    CapsuleReceived,
    CloseQuicConnection,
    ConnectStreamClosed,
    DatagramReceived,
    Effect,
    GoawayReceived,
    H3Event,
    HeadersReceived,
    LogH3Frame,
    SendQuicData,
    SettingsReceived,
    TransportDatagramFrameReceived,
    TransportStreamDataReceived,
    WebTransportStreamDataReceived,
)
from pywebtransport._protocol.utils import (
    is_bidirectional_stream,
    is_request_response_stream,
    is_unidirectional_stream,
    validate_control_stream_id,
    validate_unidirectional_stream_id,
)
from pywebtransport.constants import ErrorCodes
from pywebtransport.exceptions import ProtocolError
from pywebtransport.types import Buffer, Headers, StreamId
from pywebtransport.utils import get_logger

if TYPE_CHECKING:
    from pywebtransport._protocol.state import ProtocolState
    from pywebtransport.config import ClientConfig, ServerConfig


__all__: list[str] = []


type _RawHeaders = list[tuple[bytes, bytes]]

COLON = 0x3A
CR = 0x0D
HTAB = 0x09
LF = 0x0A
NUL = 0x0
RESERVED_SETTINGS = (0x0, 0x2, 0x3, 0x4, 0x5)
SP = 0x20
WHITESPACE = (SP, HTAB)

logger = get_logger(name=__name__)


class WebTransportH3Engine:
    """Handles WebTransport over HTTP/3 protocol parsing and encoding."""

    def __init__(self, *, is_client: bool, config: ClientConfig | ServerConfig) -> None:
        """Initialize the WebTransportH3Engine."""
        self._config = config
        self._is_client = is_client

        self._max_table_capacity = 4096
        self._blocked_streams = 16
        self._decoder = pylsqpack.Decoder(self._max_table_capacity, self._blocked_streams)
        self._encoder = pylsqpack.Encoder()

        self._settings_received = False
        self._local_control_stream_id: int | None = None
        self._local_decoder_stream_id: int | None = None
        self._local_encoder_stream_id: int | None = None
        self._peer_control_stream_id: int | None = None
        self._peer_decoder_stream_id: int | None = None
        self._peer_encoder_stream_id: int | None = None

        self._partial_frames: dict[StreamId, _PartialFrameInfo] = {}

    def cleanup_stream(self, *, stream_id: StreamId) -> None:
        """Remove any partial state associated with a closed stream."""
        self._partial_frames.pop(stream_id, None)

    def encode_capsule(self, *, stream_id: StreamId, capsule_type: int, capsule_data: bytes) -> bytes:
        """Encode a capsule into an HTTP/3 DATA frame."""
        if not is_request_response_stream(stream_id=stream_id):
            raise ProtocolError(
                message="Capsules can only be encoded for client-initiated bidirectional streams.",
                error_code=ErrorCodes.H3_STREAM_CREATION_ERROR,
            )

        capsule_buf = QuicBuffer(capacity=len(capsule_data) + 2 * UINT_VAR_MAX_SIZE)
        capsule_buf.push_uint_var(capsule_type)
        capsule_buf.push_uint_var(len(capsule_data))
        capsule_buf.push_bytes(capsule_data)

        return _encode_frame(frame_type=constants.H3_FRAME_TYPE_DATA, frame_data=capsule_buf.data)

    def encode_datagram(self, *, stream_id: StreamId, data: Buffer | list[Buffer]) -> list[Buffer]:
        """Encode a datagram payload."""
        if not is_request_response_stream(stream_id=stream_id):
            raise ProtocolError(
                message="Datagrams can only be encoded for client-initiated bidirectional streams",
                error_code=ErrorCodes.H3_STREAM_CREATION_ERROR,
            )

        header = encode_uint_var(stream_id // 4)

        if isinstance(data, list):
            return [header, *data]
        return [header, data]

    def encode_goaway_frame(self, *, last_stream_id: int = 0) -> bytes:
        """Encode an H3 GOAWAY frame."""
        buf = QuicBuffer(capacity=UINT_VAR_MAX_SIZE)
        buf.push_uint_var(last_stream_id)
        return _encode_frame(frame_type=constants.H3_FRAME_TYPE_GOAWAY, frame_data=buf.data)

    def encode_headers(self, *, stream_id: StreamId, headers: Headers, end_stream: bool = False) -> list[Effect]:
        """Encode headers and return effects to send them."""
        effects: list[Effect] = []
        raw_items = headers.items() if isinstance(headers, dict) else headers
        raw_headers: _RawHeaders = []

        for k, v in raw_items:
            key_bytes = k if isinstance(k, bytes) else k.encode("utf-8")
            val_bytes = v if isinstance(v, bytes) else v.encode("utf-8")
            raw_headers.append((key_bytes, val_bytes))

        encoder_instructions, frame_payload = self._encoder.encode(stream_id, raw_headers)
        if self._local_encoder_stream_id is not None and encoder_instructions:
            effects.append(
                SendQuicData(stream_id=self._local_encoder_stream_id, data=encoder_instructions, end_stream=False)
            )

        frame_data = _encode_frame(frame_type=constants.H3_FRAME_TYPE_HEADERS, frame_data=frame_payload)
        effects.append(SendQuicData(stream_id=stream_id, data=frame_data, end_stream=end_stream))

        effects.append(
            LogH3Frame(
                category="http",
                event="frame_created",
                data={
                    "frame_type": constants.H3_FRAME_TYPE_HEADERS,
                    "length": len(frame_payload),
                    "headers": raw_headers,
                    "stream_id": stream_id,
                },
            )
        )
        return effects

    def encode_webtransport_stream_creation(
        self, *, stream_id: StreamId, control_stream_id: StreamId, is_unidirectional: bool
    ) -> list[Effect]:
        """Return effects to signal stream creation."""
        effects: list[Effect] = []

        if is_unidirectional:
            effects.append(
                SendQuicData(
                    stream_id=stream_id, data=encode_uint_var(constants.H3_STREAM_TYPE_WEBTRANSPORT), end_stream=False
                )
            )
            effects.append(SendQuicData(stream_id=stream_id, data=encode_uint_var(control_stream_id), end_stream=False))
        else:
            frame_data = encode_uint_var(constants.H3_FRAME_TYPE_WEBTRANSPORT_STREAM) + encode_uint_var(
                control_stream_id
            )
            effects.append(SendQuicData(stream_id=stream_id, data=frame_data, end_stream=False))

        effects.append(
            LogH3Frame(category="http", event="stream_type_set", data={"new": "webtransport", "stream_id": stream_id})
        )

        if self._is_client:
            partial_info = self._get_or_create_partial_frame_info(stream_id=stream_id)
            partial_info.stream_type = constants.H3_STREAM_TYPE_WEBTRANSPORT
            partial_info.control_stream_id = control_stream_id
        return effects

    def handle_transport_event(
        self, *, event: TransportStreamDataReceived | TransportDatagramFrameReceived, state: ProtocolState
    ) -> tuple[list[H3Event], list[Effect]]:
        """Handle a translated transport event and return H3 events and effects."""
        h3_events: list[H3Event] = []
        effects: list[Effect] = []

        try:
            match event:
                case TransportStreamDataReceived(data=data, end_stream=end_stream, stream_id=stream_id):
                    if is_unidirectional_stream(stream_id=stream_id):
                        new_h3_events, new_effects = self._receive_stream_data_uni(
                            stream_id=stream_id, data=data, stream_ended=end_stream, state=state
                        )
                        h3_events.extend(new_h3_events)
                        effects.extend(new_effects)
                    else:
                        new_h3_events, new_effects = self._receive_request_data(
                            stream_id=stream_id, data=data, stream_ended=end_stream, state=state
                        )
                        h3_events.extend(new_h3_events)
                        effects.extend(new_effects)

                case TransportDatagramFrameReceived(data=data):
                    h3_events.extend(self._receive_datagram(data=data))

        except ProtocolError as exc:
            effects.append(CloseQuicConnection(error_code=exc.error_code, reason=str(exc)))
        except pylsqpack.StreamBlocked as exc:
            stream_id = int(str(exc))
            partial_info = self._partial_frames.get(stream_id)
            if partial_info is not None:
                partial_info.blocked = True
        except Exception as exc:
            effects.append(CloseQuicConnection(error_code=ErrorCodes.INTERNAL_ERROR, reason=f"H3 Engine Error: {exc}"))

        return h3_events, effects

    def initialize_connection(self) -> bytes:
        """Get the encoded SETTINGS frame payload."""
        return _encode_frame(
            frame_type=constants.H3_FRAME_TYPE_SETTINGS,
            frame_data=_encode_settings(settings=self._get_local_settings()),
        )

    def set_local_stream_ids(
        self, *, control_stream_id: StreamId, encoder_stream_id: StreamId, decoder_stream_id: StreamId
    ) -> None:
        """Set the local unidirectional stream IDs (called by Engine)."""
        validate_unidirectional_stream_id(stream_id=control_stream_id, context="Control")
        validate_unidirectional_stream_id(stream_id=encoder_stream_id, context="Encoder")
        validate_unidirectional_stream_id(stream_id=decoder_stream_id, context="Decoder")

        self._local_control_stream_id = control_stream_id
        self._local_encoder_stream_id = encoder_stream_id
        self._local_decoder_stream_id = decoder_stream_id

    def _decode_headers(self, *, stream_id: int, frame_data: bytes | None) -> tuple[_RawHeaders, list[Effect]]:
        """Decode a HEADERS frame, return headers and decoder instruction effects."""
        effects: list[Effect] = []
        decoder_instructions = b""
        try:
            if frame_data is None:
                decoder_instructions, raw_headers = self._decoder.resume_header(stream_id)
            else:
                decoder_instructions, raw_headers = self._decoder.feed_header(stream_id, frame_data)
        except pylsqpack.DecompressionFailed as exc:
            raise ProtocolError(
                message="QPACK decompression failed", error_code=ErrorCodes.QPACK_DECOMPRESSION_FAILED
            ) from exc

        if self._local_decoder_stream_id is not None and decoder_instructions:
            effects.append(
                SendQuicData(stream_id=self._local_decoder_stream_id, data=decoder_instructions, end_stream=False)
            )

        return raw_headers, effects

    @functools.cache
    def _get_local_settings(self) -> dict[int, int]:
        """Get the local HTTP/3 settings."""
        settings: dict[int, int] = {
            constants.SETTINGS_ENABLE_CONNECT_PROTOCOL: 1,
            constants.SETTINGS_H3_DATAGRAM: 1,
            constants.SETTINGS_QPACK_BLOCKED_STREAMS: self._blocked_streams,
            constants.SETTINGS_QPACK_MAX_TABLE_CAPACITY: self._max_table_capacity,
            constants.SETTINGS_WT_INITIAL_MAX_DATA: self._config.initial_max_data,
            constants.SETTINGS_WT_INITIAL_MAX_STREAMS_BIDI: self._config.initial_max_streams_bidi,
            constants.SETTINGS_WT_INITIAL_MAX_STREAMS_UNI: self._config.initial_max_streams_uni,
        }
        return settings

    def _get_or_create_partial_frame_info(self, *, stream_id: int) -> _PartialFrameInfo:
        """Get or create the partial frame info for a stream."""
        if stream_id not in self._partial_frames:
            self._partial_frames[stream_id] = _PartialFrameInfo(stream_id=stream_id)
        return self._partial_frames[stream_id]

    def _handle_control_frame(
        self, *, frame_type: int, frame_data: bytes, state: ProtocolState
    ) -> tuple[list[H3Event], list[Effect]]:
        """Handle a frame received on the control stream, return events and effects."""
        effects: list[Effect] = []
        h3_events: list[H3Event] = []
        if frame_type != constants.H3_FRAME_TYPE_SETTINGS and not self._settings_received:
            raise ProtocolError(
                message="First frame on control stream must be SETTINGS", error_code=ErrorCodes.H3_MISSING_SETTINGS
            )

        match frame_type:
            case constants.H3_FRAME_TYPE_SETTINGS:
                if self._settings_received:
                    raise ProtocolError(
                        message="SETTINGS frame received twice", error_code=ErrorCodes.H3_FRAME_UNEXPECTED
                    )
                settings = _parse_settings(data=frame_data)
                self._validate_settings(settings=settings, state=state)

                encoder_instructions = self._encoder.apply_settings(
                    max_table_capacity=settings.get(constants.SETTINGS_QPACK_MAX_TABLE_CAPACITY, 0),
                    blocked_streams=settings.get(constants.SETTINGS_QPACK_BLOCKED_STREAMS, 0),
                )

                if self._local_encoder_stream_id is not None and encoder_instructions:
                    effects.append(
                        SendQuicData(
                            stream_id=self._local_encoder_stream_id, data=encoder_instructions, end_stream=False
                        )
                    )
                self._settings_received = True
                h3_events.append(SettingsReceived(settings=settings))

            case constants.H3_FRAME_TYPE_GOAWAY:
                logger.debug("H3 GOAWAY frame received.")
                h3_events.append(GoawayReceived())

            case constants.H3_FRAME_TYPE_HEADERS:
                raise ProtocolError(
                    message="Invalid frame type on control stream", error_code=ErrorCodes.H3_FRAME_UNEXPECTED
                )

            case _:
                pass

        return h3_events, effects

    def _handle_request_frame(
        self, *, frame_type: int, frame_data: bytes | None, stream_id: int, stream_ended: bool, state: ProtocolState
    ) -> tuple[list[H3Event], list[Effect]]:
        """Handle a frame received on a request stream using a routing pattern."""
        h3_events: list[H3Event] = []
        effects: list[Effect] = []
        partial_info = self._get_or_create_partial_frame_info(stream_id=stream_id)

        match frame_type:
            case constants.H3_FRAME_TYPE_DATA:
                payload = frame_data if frame_data is not None else b""

                is_webtransport_control = partial_info.is_webtransport_control

                if not is_webtransport_control:
                    is_webtransport_control = self._is_control_stream(stream_id=stream_id, state=state)

                if is_webtransport_control:
                    if partial_info.headers_processed:
                        if payload:
                            partial_info.capsule_buffer.extend(payload)
                        if partial_info.capsule_buffer:
                            new_h3_events = self._parse_capsules(stream_id=stream_id, partial_info=partial_info)
                            h3_events.extend(new_h3_events)

                elif stream_id in state.streams:
                    session_id = state.streams[stream_id].session_id
                    session_data = state.sessions.get(session_id)
                    control_stream_id = session_id if session_data is not None else None

                    if payload:
                        if control_stream_id is not None:
                            h3_events.append(
                                WebTransportStreamDataReceived(
                                    data=payload,
                                    session_id=control_stream_id,
                                    stream_id=stream_id,
                                    stream_ended=stream_ended,
                                )
                            )
                        else:
                            raise ProtocolError(
                                message=f"Data stream {stream_id} orphaned (no control stream)",
                                error_code=ErrorCodes.INTERNAL_ERROR,
                            )
                else:
                    if payload:
                        logger.debug(
                            "Ignored DATA frame on non-WebTransport stream %d (len=%d)", stream_id, len(payload)
                        )

            case constants.H3_FRAME_TYPE_HEADERS:
                if not self._settings_received:
                    raise pylsqpack.StreamBlocked(stream_id)

                if partial_info.headers_processed:
                    raise ProtocolError(
                        message="HEADERS frame received after initial headers",
                        error_code=ErrorCodes.H3_FRAME_UNEXPECTED,
                    )

                (raw_headers, decoder_effects) = self._decode_headers(stream_id=stream_id, frame_data=frame_data)
                effects.extend(decoder_effects)

                for k, v in raw_headers:
                    if k == b":protocol" and v == b"webtransport":
                        partial_info.is_webtransport_control = True
                        break

                if self._is_client:
                    _validate_response_headers(headers=raw_headers)
                else:
                    _validate_request_headers(headers=raw_headers)

                partial_info.headers_processed = True

                frame_info = self._partial_frames.get(stream_id)
                length = 0
                if frame_data is not None:
                    length = len(frame_data)
                elif frame_info is not None and frame_info.blocked_frame_size is not None:
                    length = frame_info.blocked_frame_size
                effects.append(
                    LogH3Frame(
                        category="http",
                        event="frame_parsed",
                        data={
                            "frame_type": constants.H3_FRAME_TYPE_HEADERS,
                            "length": length,
                            "headers": raw_headers,
                            "stream_id": stream_id,
                        },
                    )
                )

                h3_events.append(
                    HeadersReceived(headers=cast(Headers, raw_headers), stream_id=stream_id, stream_ended=stream_ended)
                )

            case constants.H3_FRAME_TYPE_SETTINGS:
                raise ProtocolError(
                    message="Invalid frame type on request stream", error_code=ErrorCodes.H3_FRAME_UNEXPECTED
                )

            case (
                constants.H3_FRAME_TYPE_CANCEL_PUSH
                | constants.H3_FRAME_TYPE_MAX_PUSH_ID
                | constants.H3_FRAME_TYPE_PUSH_PROMISE
            ):
                pass

            case constants.H3_FRAME_TYPE_WEBTRANSPORT_STREAM:
                raise ProtocolError(
                    message="WT_STREAM frame (0x41) received in unexpected location",
                    error_code=ErrorCodes.H3_FRAME_ERROR,
                )

        return h3_events, effects

    def _is_control_stream(self, *, stream_id: int, state: ProtocolState) -> bool:
        """Check if the given stream ID is a WebTransport control stream (CONNECT stream)."""
        return stream_id in state.sessions

    def _parse_capsules(self, *, stream_id: int, partial_info: _PartialFrameInfo) -> list[H3Event]:
        """Parse Capsules from the accumulated data buffer in partial_info."""
        h3_events: list[H3Event] = []
        buf = QuicBuffer(data=bytes(partial_info.capsule_buffer))
        consumed = 0

        while not buf.eof():
            start_pos = buf.tell()
            try:
                capsule_type = buf.pull_uint_var()
                capsule_length = buf.pull_uint_var()

                if capsule_length > self._config.max_capsule_size:
                    raise ProtocolError(
                        message=f"Capsule length {capsule_length} exceeds limit {self._config.max_capsule_size}",
                        error_code=ErrorCodes.H3_EXCESSIVE_LOAD,
                    )

                if buf.capacity - buf.tell() < capsule_length:
                    buf.seek(start_pos)
                    break

                capsule_value = buf.pull_bytes(capsule_length)
                consumed = buf.tell()

                h3_events.append(
                    CapsuleReceived(stream_id=stream_id, capsule_type=capsule_type, capsule_data=capsule_value)
                )

            except BufferReadError:
                buf.seek(start_pos)
                break

        if consumed > 0:
            remaining_data = partial_info.capsule_buffer[consumed:]
            partial_info.capsule_buffer = bytearray(remaining_data)

        return h3_events

    def _parse_stream_data(
        self, *, stream_id: int, data_buffer: deque[Buffer], stream_ended: bool, state: ProtocolState
    ) -> tuple[list[H3Event], list[Effect]]:
        """Parse buffered data for a stream, returning events and effects."""
        h3_events: list[H3Event] = []
        effects: list[Effect] = []
        partial_info = self._get_or_create_partial_frame_info(stream_id=stream_id)

        if partial_info.blocked and partial_info.frame_type == constants.H3_FRAME_TYPE_HEADERS:
            new_h3_events, new_effects = self._handle_request_frame(
                frame_type=constants.H3_FRAME_TYPE_HEADERS,
                frame_data=None,
                stream_id=stream_id,
                stream_ended=stream_ended and not data_buffer,
                state=state,
            )
            h3_events.extend(new_h3_events)
            effects.extend(new_effects)
            partial_info.blocked = False
            partial_info.blocked_frame_size = None

        temp_data = b"".join(bytes(chunk) for chunk in data_buffer)
        consumed = 0
        buf = QuicBuffer(data=temp_data)

        while consumed < len(temp_data) or (stream_ended and consumed == len(temp_data)):
            original_consumed = consumed

            if (
                partial_info.stream_type is None
                and not partial_info.headers_processed
                and partial_info.frame_type is None
                and is_bidirectional_stream(stream_id=stream_id)
            ):
                try:
                    pos = buf.tell()
                    frame_type_check = buf.pull_uint_var()
                    if frame_type_check == constants.H3_FRAME_TYPE_WEBTRANSPORT_STREAM:
                        control_stream_id = buf.pull_uint_var()
                        validate_control_stream_id(stream_id=control_stream_id)
                        partial_info.stream_type = constants.H3_STREAM_TYPE_WEBTRANSPORT
                        partial_info.control_stream_id = control_stream_id
                        effects.append(
                            LogH3Frame(
                                category="http",
                                event="stream_type_set",
                                data={"new": "webtransport", "stream_id": stream_id},
                            )
                        )

                        h3_events.append(
                            WebTransportStreamDataReceived(
                                data=b"", session_id=control_stream_id, stream_id=stream_id, stream_ended=False
                            )
                        )
                        consumed = buf.tell()
                        continue
                    else:
                        buf.seek(pos)
                except BufferReadError:
                    buf.seek(pos)
                    break

            if partial_info.stream_type == constants.H3_STREAM_TYPE_WEBTRANSPORT:
                payload = temp_data[consumed:]
                if payload or (stream_ended and consumed == len(temp_data) and not payload):
                    control_id: int | None = partial_info.control_stream_id
                    if control_id is None:
                        stream_state = state.streams.get(stream_id)
                        session_id_fallback = stream_state.session_id if stream_state is not None else None
                        session_data_fallback = (
                            state.sessions.get(session_id_fallback) if session_id_fallback is not None else None
                        )
                        control_id = session_id_fallback if session_data_fallback is not None else None

                    if control_id is None:
                        raise ProtocolError(
                            message=f"Cannot process WT stream data for stream {stream_id} without control stream ID.",
                            error_code=ErrorCodes.INTERNAL_ERROR,
                        )

                    h3_events.append(
                        WebTransportStreamDataReceived(
                            data=payload, session_id=control_id, stream_id=stream_id, stream_ended=stream_ended
                        )
                    )
                consumed = len(temp_data)
                break
            else:
                if partial_info.frame_size is None:
                    try:
                        pos = buf.tell()
                        partial_info.frame_type = buf.pull_uint_var()
                        partial_info.frame_size = buf.pull_uint_var()
                        consumed = buf.tell()
                    except BufferReadError:
                        buf.seek(pos)
                        break

                    log_frame_size = partial_info.frame_size if partial_info.frame_size is not None else 0
                    if partial_info.frame_type == constants.H3_FRAME_TYPE_DATA:
                        effects.append(
                            LogH3Frame(
                                category="http",
                                event="frame_parsed",
                                data={
                                    "frame_type": constants.H3_FRAME_TYPE_DATA,
                                    "length": log_frame_size,
                                    "stream_id": stream_id,
                                },
                            )
                        )

                current_frame_type = partial_info.frame_type
                current_frame_size = partial_info.frame_size

                if current_frame_type is None or current_frame_size is None:
                    break

                chunk_size = min(current_frame_size, len(temp_data) - consumed)
                if current_frame_type != constants.H3_FRAME_TYPE_DATA and chunk_size < current_frame_size:
                    break

                frame_data = buf.pull_bytes(chunk_size)
                consumed = buf.tell()
                partial_info.frame_size = current_frame_size - chunk_size

                is_last_chunk = partial_info.frame_size == 0
                frame_data_to_process = frame_data if is_last_chunk else None
                if current_frame_type == constants.H3_FRAME_TYPE_DATA:
                    frame_data_to_process = frame_data

                if frame_data_to_process is not None or partial_info.blocked:
                    new_h3_events, new_effects = self._handle_request_frame(
                        frame_type=current_frame_type,
                        frame_data=frame_data_to_process,
                        stream_id=stream_id,
                        stream_ended=stream_ended and is_last_chunk and buf.eof(),
                        state=state,
                    )
                    h3_events.extend(new_h3_events)
                    effects.extend(new_effects)

                if is_last_chunk:
                    partial_info.frame_type = None
                    partial_info.frame_size = None

            if consumed == original_consumed:
                if not stream_ended:
                    logger.warning("H3 parsing stuck on stream %d", stream_id)
                break

        data_buffer.clear()
        if consumed < len(temp_data):
            data_buffer.append(temp_data[consumed:])

        return h3_events, effects

    def _receive_datagram(self, *, data: Buffer) -> list[H3Event]:
        """Parse an incoming datagram."""
        buf = QuicBuffer(data=bytes(data))
        try:
            quarter_stream_id = buf.pull_uint_var()
        except BufferReadError:
            raise ProtocolError(
                message="Could not parse quarter stream ID from datagram", error_code=ErrorCodes.H3_DATAGRAM_ERROR
            )

        stream_id = quarter_stream_id * 4
        if not is_request_response_stream(stream_id=stream_id):
            raise ProtocolError(
                message=f"Datagram received on invalid Session ID {stream_id}", error_code=ErrorCodes.H3_ID_ERROR
            )

        return [DatagramReceived(data=data[buf.tell() :], stream_id=stream_id)]

    def _receive_request_data(
        self, *, stream_id: int, data: Buffer, stream_ended: bool, state: ProtocolState
    ) -> tuple[list[H3Event], list[Effect]]:
        """Handle incoming data on a bidirectional request stream."""
        partial_info = self._get_or_create_partial_frame_info(stream_id=stream_id)
        if data:
            partial_info.buffer.append(data)
        if stream_ended:
            partial_info.ended = True

        if partial_info.blocked or (not partial_info.buffer and not partial_info.ended):
            return [], []

        h3_events, effects = self._parse_stream_data(
            stream_id=stream_id, data_buffer=partial_info.buffer, stream_ended=partial_info.ended, state=state
        )

        if partial_info.ended and not partial_info.buffer:
            stream_state = state.streams.get(stream_id)
            session_id = stream_state.session_id if stream_state is not None else None
            if session_id is not None:
                if session_id == stream_id:
                    logger.debug("CONNECT stream %d cleanly closed (FIN received)", stream_id)
                    h3_events.append(ConnectStreamClosed(stream_id=stream_id))
            self._partial_frames.pop(stream_id, None)

        return h3_events, effects

    def _receive_stream_data_uni(
        self, *, stream_id: int, data: Buffer, stream_ended: bool, state: ProtocolState
    ) -> tuple[list[H3Event], list[Effect]]:
        """Handle incoming data on a unidirectional stream."""
        partial_info = self._get_or_create_partial_frame_info(stream_id=stream_id)
        if data:
            partial_info.buffer.append(data)
        if stream_ended:
            partial_info.ended = True

        if partial_info.blocked or (not partial_info.buffer and not partial_info.ended):
            return [], []

        h3_events: list[H3Event] = []
        effects: list[Effect] = []
        temp_data = b"".join(bytes(chunk) for chunk in partial_info.buffer)
        consumed = 0
        buf = QuicBuffer(data=temp_data)

        if partial_info.stream_type is None:
            try:
                partial_info.stream_type = buf.pull_uint_var()
                consumed = buf.tell()
                stream_type = partial_info.stream_type
                uni_stream_types = (
                    constants.H3_STREAM_TYPE_CONTROL,
                    constants.H3_STREAM_TYPE_PUSH,
                    constants.H3_STREAM_TYPE_QPACK_DECODER,
                    constants.H3_STREAM_TYPE_QPACK_ENCODER,
                    constants.H3_STREAM_TYPE_WEBTRANSPORT,
                )
                if stream_type not in uni_stream_types:
                    partial_info.buffer.clear()
                    logger.warning(
                        "Received unknown unidirectional stream type %d on stream %d", stream_type, stream_id
                    )
                    return [], []

                if stream_type == constants.H3_STREAM_TYPE_CONTROL and self._peer_control_stream_id is None:
                    self._peer_control_stream_id = stream_id
                elif stream_type == constants.H3_STREAM_TYPE_QPACK_DECODER and self._peer_decoder_stream_id is None:
                    self._peer_decoder_stream_id = stream_id
                elif stream_type == constants.H3_STREAM_TYPE_QPACK_ENCODER and self._peer_encoder_stream_id is None:
                    self._peer_encoder_stream_id = stream_id

                effects.append(
                    LogH3Frame(
                        category="http",
                        event="stream_type_set",
                        data={
                            "new": {
                                constants.H3_STREAM_TYPE_CONTROL: "control",
                                constants.H3_STREAM_TYPE_PUSH: "push",
                                constants.H3_STREAM_TYPE_QPACK_ENCODER: "qpack_encoder",
                                constants.H3_STREAM_TYPE_QPACK_DECODER: "qpack_decoder",
                                constants.H3_STREAM_TYPE_WEBTRANSPORT: "webtransport",
                            }.get(stream_type, "unknown"),
                            "stream_id": stream_id,
                        },
                    )
                )
            except BufferReadError:
                return [], []

        stream_type = partial_info.stream_type

        match stream_type:
            case constants.H3_STREAM_TYPE_WEBTRANSPORT:
                if partial_info.control_stream_id is None:
                    try:
                        control_stream_id = buf.pull_uint_var()
                        validate_control_stream_id(stream_id=control_stream_id)
                        partial_info.control_stream_id = control_stream_id
                        consumed = buf.tell()
                    except BufferReadError:
                        partial_info.buffer.clear()
                        partial_info.buffer.append(temp_data[consumed:])
                        return [], []

                payload = temp_data[consumed:]
                if payload or partial_info.ended:
                    control_id = partial_info.control_stream_id
                    if control_id is None:
                        raise ProtocolError(
                            message="Cannot process WT uni stream data without control stream ID.",
                            error_code=ErrorCodes.INTERNAL_ERROR,
                        )

                    h3_events.append(
                        WebTransportStreamDataReceived(
                            data=payload, session_id=control_id, stream_ended=partial_info.ended, stream_id=stream_id
                        )
                    )
                partial_info.buffer.clear()

            case constants.H3_STREAM_TYPE_CONTROL:
                if partial_info.ended:
                    raise ProtocolError(
                        message="Closing control stream is not allowed", error_code=ErrorCodes.H3_CLOSED_CRITICAL_STREAM
                    )
                buf = QuicBuffer(data=temp_data[consumed:])
                start_pos = buf.tell()

                while not buf.eof():
                    if partial_info.frame_type is None:
                        try:
                            partial_info.frame_type = buf.pull_uint_var()
                            partial_info.frame_size = buf.pull_uint_var()
                        except BufferReadError:
                            buf.seek(start_pos)
                            break

                    needed = partial_info.frame_size
                    if needed is None:
                        break

                    if buf.capacity - buf.tell() < needed:
                        buf.seek(start_pos)
                        break

                    frame_data = buf.pull_bytes(needed)
                    frame_type = partial_info.frame_type

                    new_h3_events, new_effects = self._handle_control_frame(
                        frame_type=frame_type, frame_data=frame_data, state=state
                    )
                    h3_events.extend(new_h3_events)
                    effects.extend(new_effects)

                    partial_info.frame_type = None
                    partial_info.frame_size = None
                    start_pos = buf.tell()

                consumed += start_pos
                partial_info.buffer.clear()
                if consumed < len(temp_data):
                    partial_info.buffer.append(temp_data[consumed:])

            case constants.H3_STREAM_TYPE_QPACK_DECODER:
                if self._peer_decoder_stream_id != stream_id:
                    raise ProtocolError(
                        message="Data on unexpected QPACK decoder stream",
                        error_code=ErrorCodes.H3_STREAM_CREATION_ERROR,
                    )
                data_to_feed = temp_data[consumed:]
                try:
                    self._encoder.feed_decoder(data_to_feed)
                except pylsqpack.DecoderStreamError as exc:
                    raise ProtocolError(
                        message="QPACK decoder stream error", error_code=ErrorCodes.QPACK_DECODER_STREAM_ERROR
                    ) from exc
                partial_info.buffer.clear()

            case constants.H3_STREAM_TYPE_QPACK_ENCODER:
                if self._peer_encoder_stream_id != stream_id:
                    raise ProtocolError(
                        message="Data on unexpected QPACK encoder stream",
                        error_code=ErrorCodes.H3_STREAM_CREATION_ERROR,
                    )
                data_to_feed = temp_data[consumed:]
                unblocked_streams: set[int] = set()
                try:
                    unblocked_streams.update(self._decoder.feed_encoder(data_to_feed))
                except pylsqpack.EncoderStreamError as exc:
                    raise ProtocolError(
                        message="QPACK encoder stream error", error_code=ErrorCodes.QPACK_ENCODER_STREAM_ERROR
                    ) from exc
                partial_info.buffer.clear()

                for unblocked_stream_id in unblocked_streams:
                    unblocked_partial_info = self._partial_frames.get(unblocked_stream_id)
                    if unblocked_partial_info is not None and unblocked_partial_info.blocked:
                        unblocked_partial_info.blocked = False
                        unblocked_partial_info.blocked_frame_size = None
                        (new_h3_events, new_effects) = self._receive_request_data(
                            stream_id=unblocked_stream_id,
                            data=b"",
                            stream_ended=unblocked_partial_info.ended,
                            state=state,
                        )
                        h3_events.extend(new_h3_events)
                        effects.extend(new_effects)

        if partial_info.ended and not partial_info.buffer:
            self._partial_frames.pop(stream_id, None)

        return h3_events, effects

    def _validate_settings(self, *, settings: dict[int, int], state: ProtocolState) -> None:
        """Validate the peer's HTTP/3 settings."""
        if settings.get(constants.SETTINGS_ENABLE_CONNECT_PROTOCOL) not in (None, 1):
            raise ProtocolError(
                message="ENABLE_CONNECT_PROTOCOL setting must be 1 if present", error_code=ErrorCodes.H3_SETTINGS_ERROR
            )

        quic_supports_datagrams = state.remote_max_datagram_frame_size > 0
        if not quic_supports_datagrams and settings.get(constants.SETTINGS_H3_DATAGRAM) == 1:
            raise ProtocolError(
                message="H3_DATAGRAM requires max_datagram_frame_size transport parameter",
                error_code=ErrorCodes.H3_SETTINGS_ERROR,
            )

        webtransport_indicated = settings.get(constants.SETTINGS_ENABLE_CONNECT_PROTOCOL) == 1
        if webtransport_indicated and settings.get(constants.SETTINGS_H3_DATAGRAM) != 1:
            raise ProtocolError(
                message="WebTransport requires the H3_DATAGRAM setting", error_code=ErrorCodes.H3_SETTINGS_ERROR
            )


class _HeadersState(Enum):
    """Represent the state for tracking header frames on a stream."""

    INITIAL = 0
    AFTER_HEADERS = 1


@dataclass(kw_only=True, slots=True)
class _PartialFrameInfo:
    """Stores state for partially received H3 frames or stream data."""

    stream_id: int
    buffer: deque[Buffer] = field(default_factory=deque)
    capsule_buffer: bytearray = field(default_factory=bytearray)
    ended: bool = False
    blocked: bool = False
    blocked_frame_size: int | None = None
    frame_size: int | None = None
    frame_type: int | None = None
    stream_type: int | None = None
    control_stream_id: int | None = None
    headers_processed: bool = False
    is_webtransport_control: bool = False


def _encode_frame(*, frame_type: int, frame_data: bytes) -> bytes:
    """Encode an HTTP/3 frame."""
    frame_length = len(frame_data)
    buf = QuicBuffer(capacity=frame_length + 2 * UINT_VAR_MAX_SIZE)

    buf.push_uint_var(frame_type)
    buf.push_uint_var(frame_length)
    buf.push_bytes(frame_data)

    return buf.data


def _encode_settings(*, settings: dict[int, int]) -> bytes:
    """Encode an HTTP/3 SETTINGS frame."""
    buf = QuicBuffer(capacity=1024)
    for setting, value in settings.items():
        buf.push_uint_var(setting)
        buf.push_uint_var(value)
    return buf.data


def _parse_settings(*, data: bytes) -> dict[int, int]:
    """Parse an HTTP/3 SETTINGS frame."""
    buf = QuicBuffer(data=data)
    settings: dict[int, int] = {}
    try:
        while not buf.eof():
            setting = buf.pull_uint_var()
            value = buf.pull_uint_var()
            if setting in RESERVED_SETTINGS:
                raise ProtocolError(
                    message=f"Setting identifier 0x{setting:x} is reserved", error_code=ErrorCodes.H3_SETTINGS_ERROR
                )
            if setting in settings:
                raise ProtocolError(
                    message=f"Setting identifier 0x{setting:x} is included twice",
                    error_code=ErrorCodes.H3_SETTINGS_ERROR,
                )
            settings[setting] = value
    except BufferReadError as exc:
        raise ProtocolError(message="Malformed SETTINGS frame payload", error_code=ErrorCodes.H3_FRAME_ERROR) from exc
    return dict(settings)


def _validate_header_name(*, key: bytes) -> None:
    """Validate an HTTP header name."""
    if not key or key != key.lower():
        raise ProtocolError(
            message=f"Header name {key!r} must be lowercase and non-empty", error_code=ErrorCodes.H3_MESSAGE_ERROR
        )

    for i, c_byte in enumerate(key):
        if c_byte == COLON:
            if i == 0:
                continue
            raise ProtocolError(
                message=f"Header name {key!r} contains a non-initial colon", error_code=ErrorCodes.H3_MESSAGE_ERROR
            )

        is_tchar = (0x61 <= c_byte <= 0x7A) or (0x30 <= c_byte <= 0x39) or c_byte in b"!#$%&'*+-.^_`|~"
        if not is_tchar:
            raise ProtocolError(
                message=f"Header name {key!r} contains invalid characters", error_code=ErrorCodes.H3_MESSAGE_ERROR
            )


def _validate_header_value(*, key: bytes, value: bytes) -> None:
    """Validate an HTTP header value."""
    if not all(0x21 <= c <= 0x7E or c == SP or c == HTAB for c in value):
        raise ProtocolError(
            message=f"Header {key!r} value has forbidden characters", error_code=ErrorCodes.H3_MESSAGE_ERROR
        )
    if value and (value[0] in WHITESPACE or value[-1] in WHITESPACE):
        raise ProtocolError(
            message=f"Header {key!r} value has leading/trailing whitespace", error_code=ErrorCodes.H3_MESSAGE_ERROR
        )


def _validate_headers(
    *, headers: _RawHeaders, allowed_pseudo_headers: frozenset[bytes], required_pseudo_headers: frozenset[bytes]
) -> None:
    """Validate a list of raw HTTP headers."""
    after_pseudo_headers = False
    authority: bytes | None = None
    path: bytes | None = None
    scheme: bytes | None = None
    seen_pseudo_headers: set[bytes] = set()

    for key, value in headers:
        _validate_header_name(key=key)
        _validate_header_value(key=key, value=value)

        if key.startswith(b":"):
            if after_pseudo_headers:
                raise ProtocolError(
                    message=f"Pseudo-header {key!r} after regular headers", error_code=ErrorCodes.H3_MESSAGE_ERROR
                )
            if key not in allowed_pseudo_headers:
                raise ProtocolError(
                    message=f"Pseudo-header {key!r} not allowed", error_code=ErrorCodes.H3_MESSAGE_ERROR
                )
            if key in seen_pseudo_headers:
                raise ProtocolError(message=f"Duplicate pseudo-header {key!r}", error_code=ErrorCodes.H3_MESSAGE_ERROR)
            seen_pseudo_headers.add(key)

            match key:
                case b":authority":
                    authority = value
                case b":path":
                    path = value
                case b":scheme":
                    scheme = value
                case _:
                    pass
        else:
            after_pseudo_headers = True

    missing = required_pseudo_headers.difference(seen_pseudo_headers)
    if missing:
        missing_str = ", ".join(repr(h) for h in sorted(list(missing)))
        raise ProtocolError(message=f"Missing pseudo-headers: {missing_str}", error_code=ErrorCodes.H3_MESSAGE_ERROR)

    if scheme in (b"http", b"https"):
        if not authority:
            raise ProtocolError(
                message="Pseudo-header b':authority' cannot be empty", error_code=ErrorCodes.H3_MESSAGE_ERROR
            )
        if not path:
            raise ProtocolError(
                message="Pseudo-header b':path' cannot be empty", error_code=ErrorCodes.H3_MESSAGE_ERROR
            )


def _validate_request_headers(*, headers: _RawHeaders) -> None:
    """Validate HTTP request headers (specific for WebTransport CONNECT)."""
    _validate_headers(
        headers=headers,
        allowed_pseudo_headers=frozenset((b":method", b":scheme", b":authority", b":path", b":protocol")),
        required_pseudo_headers=frozenset((b":method", b":scheme", b":authority", b":path")),
    )


def _validate_response_headers(*, headers: _RawHeaders) -> None:
    """Validate HTTP response headers (specific for WebTransport CONNECT response)."""
    _validate_headers(
        headers=headers,
        allowed_pseudo_headers=frozenset((b":status",)),
        required_pseudo_headers=frozenset((b":status",)),
    )



"""Shared utility functions for the protocol implementation."""

from __future__ import annotations

from typing import Any

from pywebtransport.constants import MAX_STREAM_ID, ErrorCodes
from pywebtransport.exceptions import ProtocolError
from pywebtransport.types import StreamDirection, StreamId

__all__: list[str] = []

H3_ERROR_RESERVED_OFFSET = 0x21
H3_ERROR_RESERVED_MODULO = 0x1F
WT_ERROR_MAP_DIVISOR = 0x1E


def calculate_new_data_limit(
    *, current_limit: int, consumed: int, window_size: int, auto_scale: bool, force_update: bool = False
) -> int | None:
    """Calculate the new flow control limit for data."""
    if not auto_scale:
        return None

    new_limit = consumed + window_size
    threshold = 0 if force_update else window_size // 2

    if new_limit > current_limit + threshold:
        return new_limit
    return None


def calculate_new_stream_limit(
    *, current_limit: int, closed_count: int, initial_window: int, auto_scale: bool, force_update: bool = False
) -> int | None:
    """Calculate the new flow control limit for stream concurrency."""
    if not auto_scale:
        return None

    new_limit = closed_count + initial_window
    threshold = 0 if force_update else initial_window // 2

    if new_limit > current_limit and (new_limit >= current_limit + threshold):
        return new_limit
    return None


def can_receive_data_on_stream(*, stream_id: StreamId, is_client: bool) -> bool:
    """Check if the local endpoint can receive data on a given stream."""
    if is_bidirectional_stream(stream_id=stream_id):
        return True
    return (is_client and _is_server_initiated_stream(stream_id=stream_id)) or (
        not is_client and _is_client_initiated_stream(stream_id=stream_id)
    )


def can_send_data_on_stream(*, stream_id: StreamId, is_client: bool) -> bool:
    """Check if the local endpoint can send data on a given stream."""
    if is_bidirectional_stream(stream_id=stream_id):
        return True
    return (is_client and _is_client_initiated_stream(stream_id=stream_id)) or (
        not is_client and _is_server_initiated_stream(stream_id=stream_id)
    )


def get_stream_direction_from_id(*, stream_id: StreamId, is_client: bool) -> StreamDirection:
    """Determine the stream direction from its ID and the endpoint role."""
    if __debug__:
        validate_stream_id(stream_id=stream_id)

    match (
        is_bidirectional_stream(stream_id=stream_id),
        can_send_data_on_stream(stream_id=stream_id, is_client=is_client),
    ):
        case (True, _):
            return StreamDirection.BIDIRECTIONAL
        case (False, True):
            return StreamDirection.SEND_ONLY
        case (False, False):
            return StreamDirection.RECEIVE_ONLY
        case _:
            raise AssertionError("Unreachable code: Invalid stream direction logic")


def http_code_to_webtransport_code(*, http_error_code: int) -> int:
    """Map an HTTP/3 error code back to a 32-bit WebTransport application code."""
    if not (ErrorCodes.WT_APPLICATION_ERROR_FIRST <= http_error_code <= ErrorCodes.WT_APPLICATION_ERROR_LAST):
        raise ValueError("HTTP error code is not in the WebTransport application range.")

    if (http_error_code - H3_ERROR_RESERVED_OFFSET) % H3_ERROR_RESERVED_MODULO == 0:
        raise ValueError("HTTP error code is a reserved codepoint and cannot be mapped.")

    shifted = http_error_code - ErrorCodes.WT_APPLICATION_ERROR_FIRST
    return shifted - (shifted // H3_ERROR_RESERVED_MODULO)


def is_bidirectional_stream(*, stream_id: StreamId) -> bool:
    """Check if a stream is bidirectional."""
    return (stream_id & 0x2) == 0


def is_peer_initiated_stream(*, stream_id: StreamId, is_client: bool) -> bool:
    """Check if a stream was initiated by the remote peer."""
    if is_client:
        return _is_server_initiated_stream(stream_id=stream_id)
    return _is_client_initiated_stream(stream_id=stream_id)


def is_request_response_stream(*, stream_id: StreamId) -> bool:
    """Check if a stream ID is client-initiated and bidirectional."""
    return is_bidirectional_stream(stream_id=stream_id) and _is_client_initiated_stream(stream_id=stream_id)


def is_unidirectional_stream(*, stream_id: StreamId) -> bool:
    """Check if a stream is unidirectional."""
    return (stream_id & 0x2) != 0


def validate_control_stream_id(*, stream_id: StreamId) -> None:
    """Validate if an ID conforms to the H3 Session ID format (client-bidi)."""
    if not is_request_response_stream(stream_id=stream_id):
        raise ProtocolError(
            message=f"Invalid Session ID format: {stream_id} (must be client-initiated bidirectional)",
            error_code=ErrorCodes.H3_ID_ERROR,
        )


def validate_stream_id(*, stream_id: Any) -> None:
    """Validate a WebTransport stream ID."""
    if not isinstance(stream_id, int):
        raise TypeError("Stream ID must be an integer")
    if not (0 <= stream_id <= MAX_STREAM_ID):
        raise ValueError(f"Stream ID {stream_id} out of valid range")


def validate_unidirectional_stream_id(*, stream_id: StreamId, context: str = "Stream") -> None:
    """Validate that a stream ID is unidirectional."""
    if not is_unidirectional_stream(stream_id=stream_id):
        raise ProtocolError(
            message=f"{context} stream ID {stream_id} must be unidirectional.", error_code=ErrorCodes.INTERNAL_ERROR
        )


def webtransport_code_to_http_code(*, app_error_code: int) -> int:
    """Map a 32-bit WebTransport application error code to an HTTP/3 error code."""
    if not (0x0 <= app_error_code <= 0xFFFFFFFF):
        raise ValueError("Application error code must be a 32-bit unsigned integer.")

    return ErrorCodes.WT_APPLICATION_ERROR_FIRST + app_error_code + (app_error_code // WT_ERROR_MAP_DIVISOR)


def _is_client_initiated_stream(*, stream_id: StreamId) -> bool:
    """Check if a stream was initiated by the client (stream IDs are even)."""
    return (stream_id & 0x1) == 0


def _is_server_initiated_stream(*, stream_id: StreamId) -> bool:
    """Check if a stream was initiated by the server (stream IDs are odd)."""
    return (stream_id & 0x1) == 1



"""Core WebTransport connection object representing a QUIC connection."""

from __future__ import annotations

import asyncio
import uuid
import weakref
from dataclasses import dataclass
from types import TracebackType
from typing import TYPE_CHECKING, Any, Self

from pywebtransport._adapter.client import create_quic_endpoint
from pywebtransport._protocol.events import (
    ConnectionClose,
    UserConnectionGracefulClose,
    UserCreateSession,
    UserGetConnectionDiagnostics,
)
from pywebtransport.constants import ErrorCodes
from pywebtransport.events import EventEmitter
from pywebtransport.exceptions import ConnectionError, SessionError, TimeoutError
from pywebtransport.session import WebTransportSession
from pywebtransport.stream import WebTransportReceiveStream, WebTransportSendStream, WebTransportStream
from pywebtransport.types import (
    Address,
    ConnectionId,
    ConnectionState,
    EventType,
    Headers,
    SessionId,
    StreamDirection,
    StreamId,
)
from pywebtransport.utils import get_logger

if TYPE_CHECKING:
    from pywebtransport._adapter.client import WebTransportClientProtocol
    from pywebtransport._adapter.server import WebTransportServerProtocol
    from pywebtransport.config import ClientConfig, ServerConfig

    type AdapterProtocol = WebTransportServerProtocol | WebTransportClientProtocol

__all__: list[str] = ["ConnectionDiagnostics", "WebTransportConnection"]

logger = get_logger(name=__name__)

type StreamHandle = WebTransportStream | WebTransportReceiveStream | WebTransportSendStream


@dataclass(kw_only=True)
class ConnectionDiagnostics:
    """A snapshot of connection diagnostics."""

    connection_id: ConnectionId
    state: ConnectionState
    is_client: bool
    connected_at: float | None
    closed_at: float | None
    max_datagram_size: int
    remote_max_datagram_frame_size: int
    session_count: int
    stream_count: int
    active_session_handles: int
    active_stream_handles: int


class WebTransportConnection:
    """A high-level handle for a WebTransport connection over QUIC."""

    def __init__(
        self,
        *,
        config: ClientConfig | ServerConfig,
        protocol: AdapterProtocol,
        transport: asyncio.DatagramTransport,
        is_client: bool,
    ) -> None:
        """Initialize the WebTransport connection."""
        self._config = config
        self._protocol = protocol
        self._transport = transport
        self._is_client = is_client
        self._connection_id: ConnectionId = str(uuid.uuid4())
        self.events = EventEmitter(
            max_queue_size=config.max_event_queue_size,
            max_listeners=config.max_event_listeners,
            max_history=config.max_event_history_size,
        )
        self._cached_state = ConnectionState.IDLE

        self._protocol.set_status_callback(callback=self._notify_owner)

        self._session_handles: dict[SessionId, WebTransportSession] = {}
        self._stream_handles: dict[StreamId, StreamHandle] = {}

        logger.debug("WebTransportConnection %s initialized.", self.connection_id)

    @classmethod
    def accept(
        cls, *, transport: asyncio.DatagramTransport, protocol: AdapterProtocol, config: ServerConfig
    ) -> WebTransportConnection:
        """Static factory to wrap an accepted server connection."""
        connection = cls(config=config, protocol=protocol, transport=transport, is_client=False)
        return connection

    @classmethod
    async def connect(
        cls, *, host: str, port: int, config: ClientConfig, loop: asyncio.AbstractEventLoop | None = None
    ) -> WebTransportConnection:
        """Static factory to establish a client connection."""
        loop = loop or asyncio.get_running_loop()
        transport, protocol = await create_quic_endpoint(host=host, port=port, config=config, loop=loop)

        connection = cls(config=config, protocol=protocol, transport=transport, is_client=True)
        return connection

    @property
    def config(self) -> ClientConfig | ServerConfig:
        """Get the configuration associated with this connection."""
        return self._config

    @property
    def connection_id(self) -> ConnectionId:
        """Get the unique identifier for this connection."""
        return self._connection_id

    @property
    def is_client(self) -> bool:
        """Return True if this is a client-side connection."""
        return self._is_client

    @property
    def is_closed(self) -> bool:
        """Return True if the connection is closed."""
        return self.state == ConnectionState.CLOSED

    @property
    def is_closing(self) -> bool:
        """Return True if the connection is closing."""
        return self.state == ConnectionState.CLOSING

    @property
    def is_connected(self) -> bool:
        """Return True if the connection is established."""
        return self.state == ConnectionState.CONNECTED

    @property
    def local_address(self) -> Address | None:
        """Get the local address of the connection."""
        addr = self._transport.get_extra_info("sockname")
        if isinstance(addr, tuple) and len(addr) >= 2:
            return (addr[0], addr[1])
        return None

    @property
    def remote_address(self) -> Address | None:
        """Get the remote address of the connection."""
        addr = self._transport.get_extra_info("peername")
        if isinstance(addr, tuple) and len(addr) >= 2:
            return (addr[0], addr[1])
        return None

    @property
    def state(self) -> ConnectionState:
        """Get the current state of the connection."""
        return self._cached_state

    async def __aenter__(self) -> Self:
        """Enter the async context manager."""
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Exit the async context manager."""
        await self.close()

    async def close(self, *, error_code: int = ErrorCodes.NO_ERROR, reason: str = "Closed by application") -> None:
        """Immediately close the WebTransport connection."""
        if self._cached_state == ConnectionState.CLOSED:
            return

        logger.info("Closing connection %s...", self.connection_id)

        try:
            request_id, future = self._protocol.create_request()
            event = ConnectionClose(request_id=request_id, error_code=error_code, reason=reason)
            self._protocol.send_event(event=event)

            try:
                async with asyncio.timeout(delay=5.0):
                    await future
            except (asyncio.TimeoutError, asyncio.CancelledError):
                pass
            except ConnectionError as e:
                if "Connection closed" in str(e):
                    logger.debug("Connection closed while waiting for close confirmation: %s", e)
                else:
                    logger.warning("Connection error during close: %s", e)
            except Exception as e:
                logger.warning("Error during close event processing: %s", e)

        finally:
            if self.is_client:
                if not self._transport.is_closing():
                    logger.debug("Closing underlying transport for connection %s", self.connection_id)
                    self._transport.close()

            self._session_handles.clear()
            self._stream_handles.clear()
            self._cached_state = ConnectionState.CLOSED
            logger.info("Connection %s close process finished.", self.connection_id)

    async def graceful_shutdown(self) -> None:
        """Gracefully shut down the connection."""
        logger.info("Initiating graceful shutdown for connection %s...", self.connection_id)

        request_id, future = self._protocol.create_request()
        event = UserConnectionGracefulClose(request_id=request_id)
        self._protocol.send_event(event=event)

        try:
            async with asyncio.timeout(delay=5.0):
                await future
        except asyncio.TimeoutError:
            logger.warning("Timeout waiting for graceful shutdown GOAWAY confirmation.")
        except Exception as e:
            logger.warning("Error during graceful shutdown: %s", e)

        await self.close(reason="Graceful shutdown complete")

    async def create_session(self, *, path: str, headers: Headers | None = None) -> WebTransportSession:
        """Create a new WebTransport session."""
        if not self.is_client:
            raise ConnectionError("Sessions can only be created by the client.")

        request_id, future = self._protocol.create_request()
        event = UserCreateSession(request_id=request_id, path=path, headers=headers if headers is not None else {})
        self._protocol.send_event(event=event)

        try:
            session_id: SessionId = await future
        except ConnectionError:
            raise
        except asyncio.CancelledError:
            raise
        except asyncio.TimeoutError as e:
            raise TimeoutError(f"Session creation timed out: {e}") from e
        except Exception as e:
            raise SessionError(f"Session creation failed: {e}") from e

        session_handle = self._session_handles.get(session_id)
        if session_handle is None:
            logger.error("Internal error: Session handle %s missing after successful creation effect.", session_id)
            raise SessionError(f"Internal error creating session handle for {session_id}")

        return session_handle

    async def diagnostics(self) -> ConnectionDiagnostics:
        """Get diagnostic information about the connection."""
        request_id, future = self._protocol.create_request()
        event = UserGetConnectionDiagnostics(request_id=request_id)
        self._protocol.send_event(event=event)

        diag_data: dict[str, Any] = await future
        diag_data["active_session_handles"] = len(self._session_handles)
        diag_data["active_stream_handles"] = len(self._stream_handles)
        return ConnectionDiagnostics(**diag_data)

    def get_all_sessions(self) -> list[WebTransportSession]:
        """Get a list of all active session handles."""
        return list(self._session_handles.values())

    def _notify_owner(self, event_type: EventType, data: dict[str, Any]) -> None:
        """Handle high-level status events from the adapter."""
        try:
            if "connection" not in data:
                data["connection"] = weakref.proxy(self)

            if "connection_id" in data:
                data["connection_id"] = self._connection_id

            if event_type == EventType.CONNECTION_ESTABLISHED:
                self._cached_state = ConnectionState.CONNECTED
            elif event_type == EventType.CONNECTION_CLOSED:
                self._cached_state = ConnectionState.CLOSED

            if event_type in (EventType.SESSION_REQUEST, EventType.SESSION_READY):
                self._handle_session_event(event_type=event_type, data=data)

            if event_type in (
                EventType.SESSION_READY,
                EventType.SESSION_CLOSED,
                EventType.SESSION_DRAINING,
                EventType.SESSION_MAX_DATA_UPDATED,
                EventType.SESSION_MAX_STREAMS_BIDI_UPDATED,
                EventType.SESSION_MAX_STREAMS_UNI_UPDATED,
                EventType.SESSION_DATA_BLOCKED,
                EventType.SESSION_STREAMS_BLOCKED,
                EventType.DATAGRAM_RECEIVED,
            ):
                self._route_session_event(event_type=event_type, data=data)
            elif event_type in (EventType.STREAM_OPENED, EventType.STREAM_CLOSED):
                self._handle_stream_event(event_type=event_type, data=data)

            self.events.emit_nowait(event_type=event_type, data=data)

        except Exception as e:
            logger.error("Error during owner notification callback: %s", e, exc_info=True)

    def _handle_session_event(self, *, event_type: EventType, data: dict[str, Any]) -> None:
        """Create or update session handles based on events."""
        session_id = data.get("session_id")
        if session_id is None:
            return

        create_handle = (not self.is_client and event_type == EventType.SESSION_REQUEST) or (
            self.is_client and event_type == EventType.SESSION_READY
        )

        if create_handle and session_id not in self._session_handles:
            path = data.get("path")
            headers = data.get("headers")

            if path is not None and headers is not None:
                session = WebTransportSession(connection=self, session_id=session_id, path=path, headers=headers)
                self._session_handles[session_id] = session
                logger.debug("Created session handle for %s", session_id)
                data["session"] = session
            else:
                logger.error("Missing metadata for session handle creation %s", session_id)

    def _route_session_event(self, *, event_type: EventType, data: dict[str, Any]) -> None:
        """Route events to existing session handles."""
        session_id = data.get("session_id")
        if session_id is None:
            return

        session = self._session_handles.get(session_id)
        if session is not None:
            data["session"] = session
            session.events.emit_nowait(event_type=event_type, data=data)

            if event_type == EventType.SESSION_CLOSED:
                self._session_handles.pop(session_id, None)
                asyncio.create_task(coro=session.events.close())

    def _handle_stream_event(self, *, event_type: EventType, data: dict[str, Any]) -> None:
        """Manage stream handles."""
        stream_id = data.get("stream_id")
        if stream_id is None:
            return

        if event_type == EventType.STREAM_OPENED:
            session_id = data.get("session_id")
            direction = data.get("direction")

            if session_id is not None and direction is not None and stream_id not in self._stream_handles:
                session = self._session_handles.get(session_id)
                if session is not None:
                    handle_class: type[StreamHandle]
                    match direction:
                        case StreamDirection.BIDIRECTIONAL:
                            handle_class = WebTransportStream
                        case StreamDirection.SEND_ONLY:
                            handle_class = WebTransportSendStream
                        case StreamDirection.RECEIVE_ONLY:
                            handle_class = WebTransportReceiveStream
                        case _:
                            logger.error("Unknown stream direction: %s", direction)
                            return

                    new_stream = handle_class(session=session, stream_id=stream_id)
                    self._stream_handles[stream_id] = new_stream
                    data["stream"] = new_stream

                    session.events.emit_nowait(event_type=event_type, data=data)
                else:
                    logger.warning("Session %s not found for stream %d", session_id, stream_id)

        elif event_type == EventType.STREAM_CLOSED:
            stream = self._stream_handles.pop(stream_id, None)
            if stream is not None:
                data["stream"] = stream
                stream.events.emit_nowait(event_type=event_type, data=data)
                asyncio.create_task(coro=stream.events.close())

    def __repr__(self) -> str:
        """Provide a developer-friendly representation."""
        return f"<WebTransportConnection id={self.connection_id} state={self._cached_state} client={self.is_client}>"
